#include <math.h>
#include <cblas.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
//#include "clapack.h"
#include <stdbool.h>

#include "cluster.h"
#include "ds.h"

/*
#define COMPLETE 0
#define UPGMA 1
#define AVERAGE 1
#define SINGLE 2
*/


double sumsq(double *matrix, int rows, int cols, int *clusters, int k)
{
  double sswc = 0;
  double sumd, sumd2, clustsize;
  int i, j, c;
  
  /*for (i = 0; i < rows; i++)
  {
    for (j = 0; j < cols; j++)  fprintf(stderr, "%10.6f", matrix[i * cols + j]);
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n\n"); */

  for (i = 0; i < cols; i++)
  {
    for (c = 0; c < k; c++)
    {
      clustsize = 0;
      sumd = 0;
      sumd2 = 0;
      
      for (j = 0; j < rows; j++)
      {
        if (clusters[j] == c)
        {
          clustsize++;
          sumd += matrix[j * cols + i];
          sumd2 += (matrix[j * cols + i] * matrix[j * cols + i]);
        }
      } // end for j
      
      //fprintf(stderr, "For cluster %d: %f\n", c, sumd2 - ((sumd * sumd) / clustsize));
      if (clustsize > 0)  sswc += (sumd2 - ((sumd * sumd) / clustsize));
    } //end for c
  } // end for i

  return sswc;
}

double* eucdist(double *distmat, double *vectors, int nvals, int nvects)
{
  // dimensions: number of vectors (cols, eigen vectors)
  // number of values: number of values in each EV
  // loop over all pairs of values, get all distance pairs
  // to calculate each distance, loop over all dimensions to get sum
  
  int i, j, k;
  
  for (i = 0; i < nvals; i++)
  {
    distmat[i * nvals + i] = 0;
    for (j = 0; j < i; j++)
    {
      distmat[i * nvals + j] = 0;
      
      for (k = 0; k < nvects; k++)  
	      //distmat[i * nvals + j] += pow(vectors[i * nvals + k] - vectors[j * nvals + k], 2);
        distmat[i * nvals + j] += pow(vectors[i * nvects + k] - vectors[j * nvects + k], 2);
        
      distmat[i * nvals + j] = sqrt(distmat[i * nvals + j]);
      distmat[j * nvals + i] = distmat[i * nvals + j];
    }
  }
  
  return distmat;
}

double* delrowcol(double *matrix, int dim, int rowcolidx)
{
  int i;
  int *shiftby = (int*)malloc(dim * dim * sizeof(int));
  
  /* Delete one value from each row first */
  for (i = 0; i < (dim * dim); i++)
    shiftby[i] = (i + dim - rowcolidx - 1) / dim;
  
  for (i = 0; i < (dim * dim); i++)
    matrix[i - shiftby[i]] = matrix[i];
  
  
  free(shiftby);
  
  /* Delete whole row */
  for (i = rowcolidx * (dim - 1); i < ((dim - 1) * (dim - 1)); i++)
    matrix[i] = matrix[i + dim - 1];
  
  return matrix;
}

// for now, just UPGMA... implement complete linkage, single linkage too
void hierclust(double *dmat, int rows, int cols, int *clusters, int k, HCLMETHOD method)
{
  int i, j;
  int mini, minj;
  double propi, propj;
  int nclusts = rows;
  double mindist;
  //double *dmat = (double*)malloc(rows * rows * sizeof(double));
  //double *tipdists = (double*)malloc(rows * sizeof(double));
  size_t *clustsizes = (size_t*)malloc(rows * sizeof(size_t));
    
  //eucdist(dmat, matrix, rows, cols);

  
  for (i = 0; i < rows; i++)  
  {
    clusters[i] = i;
    clustsizes[i] = 1;
  }
  
  
  while (nclusts > k)
  {
    mindist = -1;
    for (i = 0; i < nclusts; i++)
    {
      for (j = 0; j < i; j++)
      {
        if ((mindist < 0) || (dmat[i * nclusts + j] < mindist))
        {
          mindist = dmat[i * nclusts + j];
          mini = i;
          minj = j;
        }
        
      }        
    }
        
    propi = (double)(clustsizes[mini]) / (clustsizes[mini] + clustsizes[minj]);
    propj = 1.0 - propi;
    
    
    for (i = 0; i < nclusts; i++)
    {
      if (i != minj)
      {
        switch (method)
        {
          case AVERAGE:
            dmat[i * nclusts + minj] = (dmat[mini * nclusts + i] * propi) + (dmat[minj * nclusts + i] * propj);
            break;
          case COMPLETE:
            dmat[i * nclusts + minj] = MAX(dmat[mini * nclusts + i], dmat[minj * nclusts + i]);
            break;
          case SINGLE:
            dmat[i * nclusts + minj] = MIN(dmat[mini * nclusts + i], dmat[minj * nclusts + i]);
            break;
          default:
            fprintf(stderr, "Error! Invalid clustering method. Must specify SINGLE, COMPLETE, or AVERAGE.\n");
            goto hierclustend;
            break;
        }
        
        dmat[minj * nclusts + i] = dmat[i * nclusts + minj];
      }
    }
     
    delrowcol(dmat, nclusts, mini);
    clustsizes[minj] += clustsizes[mini];
    
    for (i = mini; i < (nclusts - 1); i++)  clustsizes[i] = clustsizes[i + 1];
    
    for (i = 0; i < rows; i++)
    {
      if (clusters[i] == mini)  clusters[i] = minj;
      
      else if (clusters[i] > mini)  clusters[i]--;
    }
 
    nclusts--;
  }
  
  //for (i = 0; i < rows; i++)  printf(" %d", clusters[i]);
  
  //printf("\n");
  
  hierclustend:
    //free(dmat);
  free(clustsizes);
}

/* buffer can be NULL */
/*
void readmatrix(double *matrix, int N, FILE *matfile, char *buffer)
{
   int bufnull = 0;
   int valcount = 0;
   int maxcount = N * N;   
   char *token;
   
   if (!buffer)
   {
     bufnull = 1;
     buffer = (char*)malloc(BUFSIZ * sizeof(char));
   }
   
   while (fgets(buffer, BUFSIZ, matfile))
   {
     token = strtok(buffer, " \n\r\t");
     while (token)
     {
       if (valcount >= maxcount)
       {
         fprintf(stderr, "Error! Too many values in input file (max is %d; last val %f)!\n", maxcount, matrix[valcount-1]);
         exit(1);
       }
       
       matrix[valcount] = atof(token);
       valcount++;
       token = strtok(NULL, " \n\r\t");
     }
   
   }
   
   if (bufnull)  free(buffer);
}*/

void transpose(double *matrix, int N)
{
  int i, j;
  double temp;
  
  for (i = 0; i < N; i++)
  {
    for (j = 0; j < i; j++)
    {
      temp = matrix[i * N + j];
      matrix[i * N + j] = matrix[j * N + i];
      matrix[j * N + i] = temp;
    }
  }
}


/**
 *  Params:
 *    JOBVL, JOBVR: if 'N' left/right eigenvectors not computed; if 'V' they are
 *    N: order of the matrix
 *    A: Array of values; dim=(LDA,N)
 *    LDA: Leading dimension of A, LDA >= max(1,N)
 *    WR, WI: real and imaginary parts of eigenvalues (dim=N)
 *    VL: If JOBVL = 'V', stores left eigenvectors; dim=(LDVL, N)
 *    LDVL: Leading dimension of VL
 *    VR: If JOBVR = 'V', stores right eigenvectors; dim=(LDVR, N)
 *    LDVR: Leading dimension of VR
 *    WORK: Workspace, dim=max(1,LWORK)
 *    LWORK: Leading dimension of WORK, LWORK >= max(1, 3*N) (if both JOBVL and
 *           JOBVR = 'V', LWORK >= max(1, 4*N), should normally be larger for 
 *           good performance)
 *    
 *  Return: 0 on successful exit
 *          < 0, illegal val; -i indicates that ith arg was illegal
 *          > 0, failed QR; i indicates that WR and WI [i+1:N] have converged
 */
static long dgeev(char JOBVL, char JOBVR, int N, double *A, int LDA, double *WR,
                  double *WI, double *VL, int LDVL, double *VR, int LDVR, 
                  double *WORK, int LWORK)
{
  extern void dgeev_(char*, char*, int*, double*, int*, double*, double*, double*, 
                     int*, double*, int*, double*, int*, long*);
  long info;
  
  dgeev_(&JOBVL, &JOBVR, &N, A, &LDA, WR, WI, VL, &LDVL, VR, &LDVR, WORK, 
         &LWORK, &info);
  
  return info;
}

int geteigvecs(double *distances, double *eigvr, double *eigvec, int ngenes)
{
  double *Amat, *Dmat, *Lmat; // *Xmat, *Ymat, *eigdists;
  graph *affingraph;
  char **labels;
  bool connected = false;
  int i, j, info;
  double sigma2 = 1.;
  double x;
  //double *eigvr, 
  double *eigvi, *workspace;
  double rowsum;
  double d;
  int nth = (int)(log(ngenes) + 0.5) + 1;
  
  //FILE *temp;
  
  double *scales = (double*)malloc(ngenes * sizeof(double));
  //printf("Scales:");
  //sigma2 = calcscales(scales, distances, ngenes, nth);
  //printf("\nsigma2: %.6f\n", sigma2);
  //sigma2 *= sigma2;
  
  Amat = (double*)malloc(ngenes * ngenes * sizeof(double));
  Dmat = (double*)malloc(ngenes * sizeof(double));
  
  affingraph = (graph*)malloc(sizeof(graph));
  labels = (char**)malloc(ngenes * sizeof(char*));
  
  for (i = 0; i < ngenes; i++)
  {
    labels[i] = (char*)malloc(3 * sizeof(char));
    sprintf(labels[i], "%02d", i);
  }
  
  while (! connected)
  {
    sigma2 = calcscales(scales, distances, ngenes, nth);
    sigma2 *= sigma2;
    //temp = fopen("Amat", "w");
    for (i = 0; i < ngenes; i++)  Dmat[i] = 0.;

    for (i = 0; i < ngenes; i++)
    {
      for (j = 0; j < i; j++)
      {
        d = distances[i * ngenes + j];
        if (d > scales[i] && d > scales[j])
        {
          Amat[i * ngenes + j] = 0;
          Amat[j * ngenes + i] = 0;
        }
        else
        {
          x = exp(- pow(d, 2) / 
          (scales[i] * scales[j])); 
          //sigma2);
          Amat[i * ngenes + j] = x;
          Amat[j * ngenes + i] = x;
          Dmat[i] += x;
          Dmat[j] += x;
          //fprintf(temp, "%s %s pp %f\n", labels[i], labels[j], Amat[i * ngenes + j]);
        }
      }
    }
    
    buildgraph(affingraph, Amat, labels, ngenes, true);
    connected = isconnected(affingraph);
    freegraph(affingraph);
    
    //printf("connected: %d ; nvertices: %d ; nedges: %d ;  nth: %d\n", connected, affingraph->nvertices, affingraph->nedges, nth);
    //fclose(temp);
    nth++;
  }
  
  free(affingraph);
  
  for (i = 0; i < ngenes; i++)  free(labels[i]);
  free(labels);

  Lmat = (double*)malloc(ngenes * ngenes * sizeof(double));
  
  for (i = 0; i < ngenes; i++)
  {
    for (j = 0; j < ngenes; j++)
    {
      x = Amat[i * ngenes + j] / sqrt(Dmat[i]);
      Lmat[i * ngenes + j] = x / sqrt(Dmat[j]);
    }
  }
  
  transpose(Lmat, ngenes);
  //eigvr = (double*)malloc(ngenes * sizeof(double));
  eigvi = (double*)malloc(ngenes * sizeof(double));
  workspace = (double*)malloc(8 * ngenes * sizeof(double));
  
  info = dgeev('N', 'V', ngenes, Lmat, ngenes, eigvr, eigvi, NULL, 1, eigvec, ngenes, workspace, 8 * ngenes);
  evsort(eigvr, eigvec, ngenes, 0, ngenes - 1);
  
  /*for (i = 0; i < ngenes; i++)  printf(" %.6f", eigvr[i]);
  printf("\n");*/
    
  free(scales);
  free(workspace);
  free(eigvi);
  //free(eigvr);
  free(Lmat);
  free(Dmat);
  free(Amat);
  
  return info;
}

double calcscales(double *scales, double *distances, int dim, int nth)
{
  int i, j;
  //int nth = (int)(log(dim) + 0.5) + 1;
  double scalesum = 0;
  double *nodedists = (double*)malloc(dim * sizeof(double));
  
  for (i = 0; i < dim; i++)
  {
    for (j = 0; j < dim; j++)
      nodedists[j] = distances[i * dim + j];
    
    qsort(nodedists, dim, sizeof(double), cmpdp);
    scales[i] = nodedists[nth];
    scalesum += scales[i];
    
    //printf(" %.6f", scales[i]);
  }
  
  free(nodedists);
  return scalesum / dim;
}

static int cmpdp(const void *p1, const void *p2)
{
  double d1, d2;
  
  d1 = *(double *)p1;
  d2 = *(double *)p2;
  
  if (d1 == d2)  return 0;
  
  return d1 > d2 ? 1 : -1;
}

void evsort(double *eigval, double *eigvec, int dim, int start, int end)
{
  int i;
  int pivot;
  int placeholder;
  double pivotval;
  double temp;
  if (start >= end)  return;
  
  pivot = start + (rand() % (end - start + 1));
  pivotval = eigval[pivot];
  eigval[pivot] = eigval[end];
  eigval[end] = pivotval;
  swaprows(eigvec, dim, pivot, end);
  
  placeholder = start;
  for (i = start; i < end; i++)
  {
    if (eigval[i] >= pivotval)
    {
      temp = eigval[placeholder];
      eigval[placeholder] = eigval[i];
      eigval[i] = temp;
      swaprows(eigvec, dim, i, placeholder);
      placeholder++;
    }
  }
  
  temp = eigval[placeholder];
  eigval[placeholder] = eigval[end];
  eigval[end] = temp;
  swaprows(eigvec, dim, placeholder, end);
  
  evsort(eigval, eigvec, dim, start, placeholder - 1);
  evsort(eigval, eigvec, dim, placeholder + 1, end);
}

void swaprows(double *matrix, int cols, int rowidxA, int rowidxB)
{
  double temp;
  int i;
  
  for (i = 0; i < cols; i++)
  {
    temp = matrix[rowidxA * cols + i];
    matrix[rowidxA * cols + i] = matrix[rowidxB * cols + i];
    matrix[rowidxB * cols + i] = temp;
  }
}

double esclust(double *eigvec, int *clusters, int ngenes, int k, HCLMETHOD method)
{
  double *Xmat, *Ymat;
  double *eigdists;
  int i, j;
  double rowsum;
  double chidx, ssw, ssb;
  
  Ymat = (double*)malloc(k * ngenes * sizeof(double));
  Xmat = (double*)malloc(k * ngenes * sizeof(double));
  
  for (i = 0; i < ngenes; i++)
  {
    rowsum = 0;
    for (j = 0; j < k; j++)
    {
      /* Xmat contains first k eigenvectors, transposed into columns */
      Xmat[i * k + j] = eigvec[j * ngenes + i];
      rowsum += pow(Xmat[i * k + j], 2);
    }
    
    rowsum = sqrt(rowsum);
    
    for (j = 0; j < k; j++)
    {
      Ymat[i * k + j] = Xmat[i * k + j] / rowsum;
    }
  }
    
  eigdists = (double*)malloc(ngenes * ngenes * sizeof(double));
  eucdist(eigdists, Ymat, ngenes, k);
  hierclust(eigdists, ngenes, k, clusters, k, method);
  
  //fprintf(stderr, " CH idx / ssw of Ymat: %f ", 
  chidx = chindexeuc(clusters, Ymat, ngenes, k, k, &ssw, &ssb);
  //fprintf(stderr, "%f %f\n", ssw, ssb);
  //fprintf(stderr, "chidx: %f\n", chidx);
  
  //ssw = sumsq(Ymat, ngenes, k, clusters, k);
  
  free(Xmat);
  free(Ymat);

  return ssw;
}

void writeeigvects(FILE *outfile, double *eigvec, int ngenes, int k)
{
  double *Xmat;
  int i, j;
  double rowsum;

  Xmat = (double*)malloc(k * ngenes * sizeof(double));
  
  for (i = 0; i < ngenes; i++)
  {
    rowsum = 0;
    for (j = 0; j < k; j++)
    {
      /* Xmat contains first k eigenvectors, transposed into columns */
      Xmat[i * k + j] = eigvec[j * ngenes + i];
      rowsum += pow(Xmat[i * k + j], 2);
    }
    
    rowsum = sqrt(rowsum);
    
    for (j = 0; j < k; j++)
      fprintf(outfile, " %10.6g", (Xmat[i * k + j] / rowsum)); 
    
    fprintf(outfile, "\n");
  }
    
  free(Xmat);
}

int eigengap(double* eigvals, int nvals)
{
  int i;
  int maxgapidx = 1;
  double gap;
  double maxgap;
  
  if (nvals < 2)  return -1;
  
  maxgap = eigvals[maxgapidx - 1] - eigvals[maxgapidx];
  
  for (i = 2; i < (nvals - 1); i++)
  {
    gap = eigvals[i - 1] - eigvals[i];
    if (gap > maxgap)
    {
      maxgap = gap;
      maxgapidx = i;
    }
  }
  
  return maxgapidx;
}

double chindexeuc(int *clusters, double *matrix, int rows, int cols, int k, double *Wk, double *Bk)
{
  int i, j, c;
  double ssbt = 0., sswt = 0.;
  double x2t, xt, xct, te;
  int nc, idx, ng;
  
  *Wk = 0.;
  *Bk = 0.;
  
  //for (i = 0; i < ngenes; i++)
  for (i = 0; i < cols; i++)
  {
    x2t = xt = ssbt = sswt = 0.;
    ng = 0;
    
    for (c = 0; c < k; c++)
    {
      xct = 0.;
      nc = 0;

      //for (j = 0; j < ngenes; j++)
      for (j = 0; j < rows; j++)     
      {
        if (clusters[j] == c)
        {
          //idx = i * ngenes + j;
          idx = j * cols + i;

         /*x2t += distances[idx] * distances[idx];
          xct += distances[idx];*/
          x2t += matrix[idx] * matrix[idx];
          xct += matrix[idx];
          nc++;
          ng++;
        }
      }
      
      //if (nc == 0)  printf("PROBLEM: nc for this cluster/column is 0!\n");
      
      if (nc)
      {
        xt += xct;
        te = (xct * xct)/nc;
        ssbt += te;
        sswt -= te;
      }
    }
    
    if (ng != rows)  printf("Valid distances in column: %d genes: %d\n", ng, rows);
    ssbt -= (xt * xt)/ng;
    sswt += x2t;
    
    *Wk += sswt;
    *Bk += ssbt;
  }
  
  //fprintf(stderr, "Wk, Bk inside chindex is %f %f; ", *Wk, *Bk);
  return (*Bk / (k - 1)) / (*Wk / (rows - k));//(ngenes - k));
}

