#include <stdlib.h>
#include <stdio.h>

#include "ds.h"

/* Function prototypes for internal use only */
graph* testgraph();
graph* testgraph2();
graph* testgraph3();
//graph* testgraph4();
vertex* initvertex(vertex*, int);
void freevertex(vertex*);
edge* initedge(edge*, double, vertex*, vertex*);
edge* addincidentedge(vertex*, edge*);
int cmpedges(const void*, const void*);
int cmpvertices(const void*, const void*);
void sortedges(graph*);

vertex* initvertex(vertex *v, int vindex)
{
  v->maxedges = INITITEMS;
  v->incidentedges = (edge**)malloc(INITITEMS * sizeof(edge*));
  v->degree = 0;
  v->visited = FALSE;
  v->vindex = vindex;

  return v;
}

void freevertex(vertex *v)
{
  free(v->incidentedges);
}

edge* initedge(edge *e, double weight, vertex *u, vertex *v)
{
  e->weight = weight;
  e->vertices[0] = u;
  e->vertices[1] = v;
  e->traversed = FALSE;

  return e;
}

graph* initgraph(graph *g, int nvertices, int nedges)
{
  g->vertices = (vertex**)malloc(nvertices * sizeof(vertex*));
  g->edgelist = (edge**)malloc(nedges * sizeof(edge*));
  g->labels = (char**)malloc(nvertices * sizeof(char*));
  
  g->nvertices = g->nedges = 0;

  g->maxvertices = nvertices;
  g->maxedges = nedges;

  return g;
}

void freegraph(graph *g)
{
  int i;
  for (i = 0; i < g->nedges; i++) free(g->edgelist[i]);
  for (i = 0; i < g->nvertices; i++)
  {
    freevertex(g->vertices[i]);
    free(g->vertices[i]);
  }

  free(g->labels);
  free(g->vertices);
  free(g->edgelist);
}

vertex* addvertexgraph(graph *g, char *label)
{
  vertex *v = (vertex*)malloc(sizeof(vertex));
  initvertex(v, g->nvertices);

  if (g->nvertices >= g->maxvertices)
  {
    g->maxvertices *= 2;
    g->vertices = (vertex**)realloc(g->vertices, g->maxvertices * sizeof(vertex*));
    g->labels = (char**)realloc(g->labels, g->maxvertices * sizeof(char*));
  }

  g->vertices[g->nvertices] = v;
  g->labels[g->nvertices] = label;
  g->nvertices++;

  return v;
}

edge* addedgegraph(graph *g, vertex *u, vertex *v, double weight)
{
  edge *e = (edge*)malloc(sizeof(edge));
  initedge(e, weight, u, v);
  addincidentedge(u, e);
  addincidentedge(v, e);

  if (g->nedges >= g->maxedges)
  {
    g->maxedges *= 2;
    g->edgelist = (edge**)realloc(g->edgelist, g->maxedges * sizeof(edge*));
  }

  g->edgelist[g->nedges] = e;
  g->nedges++;

  return e;

}

edge* addincidentedge(vertex *v, edge *e)
{
  if (v->degree >= v->maxedges)
  {
    v->maxedges *= 2;
    v->incidentedges = (edge**)realloc(v->incidentedges, v->maxedges * sizeof(edge*));
  }

  v->incidentedges[v->degree] = e;
  v->degree++;

  return e;
}

void sortedges(graph *g)
{
  qsort(g->edgelist, g->nedges, sizeof(edge*), cmpedges);
}


/**
 * Accessory function for removeminvertset; see below.
 */
// int countdegree(vertex *v, bool *setmembers)
// {
//   int i, count;
//   edge *e;
//   vertex *neighbour;
//   
//   for (i = 0, count = 0; i < v->degree; i++)
//   {
//     e = v->incidentedges[i];
//     if (e->vertices[0] == v)  neighbour = e->vertices[1];
//     
//     else if (e->vertices[1] == v) neighbour = e->vertices[0];
//     
//     else
//     {
//       fprintf(stderr, "Error! Adjacency not found!\n");
//       exit(1);
//     }
//     
//     if (setmembers[neighbour->vindex]) count += 1;
//   }
//   
//   return count;
// }

/*graph* buildgraph(graph *g, double *distmat, char **labels, int nvertices)
{
  return buildgraph(g, distmat, labels, nvertices, false);
}*/

/**
 *  Note that if a distance value less than 0 is found, the edge is not added.
 * For affinity matrices, edges are only added for values strictly greater than 0.
 */
graph* buildgraph(graph *g, double *distmat, char **labels, int nvertices, bool isaffinity)
{
  int i, j;
  double dist; 
  
  initgraph(g, nvertices, (nvertices * (nvertices-1))/2);

  for (i = 0; i < nvertices; i++)
  {
    addvertexgraph(g, labels[i]);
    for (j = 0; j < i; j++)  
    {
      dist = distmat[i * nvertices + j];
      
      if (isaffinity && dist > 0)  
        addedgegraph(g, g->vertices[i], g->vertices[j], dist);
        
      else if (! isaffinity && dist >= 0)  addedgegraph(g, g->vertices[i], g->vertices[j], dist);
    }
    
  }
  //g->labels = labels;
}


int cmpedges(const void *p1, const void *p2)
{

  if ((*(edge* const*)p1)->weight > (*(edge* const*)p2)->weight) return 1;
  
  if ((*(edge* const*)p1)->weight < (*(edge* const*)p2)->weight) return -1;

  return 0;

}

int cmpvertices(const void *p1, const void *p2)
{
  if (((vertex*)p1)->rank > ((vertex*)p2)->rank)  return 1;
  if (((vertex*)p1)->rank < ((vertex*)p2)->rank)  return -1;
  
  return 0;
}

bool isconnected(graph *g)
{
  int i;
  linkedlist *verts;
  vertex *u, *v;
  edge *e;
  
  if (g->nvertices <= 1)  return true;

  verts = (linkedlist*)malloc(sizeof(linkedlist));
  initlinkedlist(verts);
  
  int llsize = 1;
  enqueuelinkedlist(verts, g->vertices[0]);
  
  
  for (i = 0; i < g->nvertices; i++)
    g->vertices[i]->visited = 0;
  
  
  while (llsize > 0)
  {
    v = (vertex*)dequeuelinkedlist(verts);
    llsize--;
    v->visited = 1;
    for (i = 0; i < v->degree; i++)
    {
      e = v->incidentedges[i];
      u = ((v == e->vertices[0]) ? 
           e->vertices[1] : 
           e->vertices[0]);
           
      if (! u->visited)
      {
        enqueuelinkedlist(verts, u);
        llsize++;
      }
    }
  }
  
  freelinkedlist(verts);
  free(verts);
  
  for (i = 0; i < g->nvertices; i++)
    if (! g->vertices[i]->visited)  return false;

  return true;
}

/**
 * Calculate distances between src and all other vertices, ignoring edge 
 * weights if ignoreweights == TRUE.
 */
vertex** Dijkstragraph(graph *g, vertex *src, vertex **paths, double *distances, double maxdist, bool ignoreweights)
{
  int i, j;
  double newdist;
  vertex *target, *u, *v;
  edge *e;
  //vertqueue *vQ;
  heap *vheap;
  
  vheap = (heap*)malloc(sizeof(heap));
  initheap(vheap, cmpvertices);
  //vQ = (vertqueue*)malloc(sizeof(vertqueue));
  //initvertqueue(vQ);
  
  for (i = 0; i < g->nvertices; i++)
  {
    g->vertices[i]->visited = FALSE;
    distances[i] = -1; //g->vertices[i]->vindex] = -1;
    //addvertqueue(vQ, g->vertices[i], g->vertices[i]->degree);
  }
  
  /*printf("Testing vertqueue:\n");
  while(v = remvertqueue(vQ))
  {
    printf("  %s: degree = %d\n", g->labels[v->vindex], v->degree);
  }*/
  src->rank = 0;
  addheap(vheap, src);
  distances[src->vindex] = 0;
  paths[src->vindex] = NULL;
  
  /* While vQ is not empty... breadth-first based distance calc */
  while (v = (vertex*)removeheap(vheap))//remvertqueue(vQ))
  {
    if (!v->visited)
    {
      qsort(v->incidentedges, v->degree, sizeof(edge*), cmpedges);
      
      for (i = 0; i < v->degree; i++)
      {
        e = v->incidentedges[i];
      
        if (e->vertices[0] == v)  u = e->vertices[1];
      
        else if (e->vertices[1] == v)  u = e->vertices[0];
      
        else 
        {
          fprintf(stderr, "Error! Adjacency not found!\n");
          exit(1);
        }
        
        if (!u->visited)
        {
          newdist = distances[v->vindex];
          if (ignoreweights)  newdist++;
          else  newdist += e->weight;
      
          if (newdist <= maxdist && (distances[u->vindex] < 0 || newdist < distances[u->vindex]))
          {
            distances[u->vindex] = newdist;
            paths[u->vindex] = v;
            u->rank = newdist;
            addheap(vheap, u);
            //addvertqueue(vQ, u, newdist);
          }
        }
      }
      
      v->visited = TRUE;
    }
  }
    // sort incident edges!
  //freevertqueue(vQ);
  //free(vQ);
  
  freeheap(vheap);
  free(vheap);
  
  return paths;
}

/* closeness, eccentricity optional */
void centralitiesgraph(graph *g, double *betweenness, double *closeness, double *eccentricity)
{
  int i, j, size;
  double newdist;
  vertex *target, *u, *v, *src;
  edge *e;
  heap *vheap;
  linkedlist *stack;
  vector **predecessors = (vector**)malloc(g->nvertices * sizeof(vector*));
  double *distances = (double*)malloc(g->nvertices * sizeof(double));
  double *sigma = (double*)malloc(g->nvertices * sizeof(double));
  double *delta = (double*)malloc(g->nvertices * sizeof(double));
  
  
  
  stack = (linkedlist*)malloc(sizeof(linkedlist));
  initlinkedlist(stack);  
  
  vheap = (heap*)malloc(sizeof(heap));
  initheap(vheap, cmpvertices);
  
  for (i = 0; i < g->nvertices; i++)
  {
    betweenness[i] = 0;
    if (closeness)  closeness[i] = 0;
    if (eccentricity)  eccentricity[i] = -1;
    predecessors[i] = (vector*)malloc(sizeof(vector));
    initvector(predecessors[i]);
  }
  
  for (i = 0; i < g->nvertices; i++)
  {
    src = g->vertices[i];
    
    // make sure heap, stack are empty?
    //while (v = (vertex*)poplinkedlist(stack)) ; /* Make sure stack is empty */
    
    for (j = 0; j < g->nvertices; j++)
    {
      g->vertices[j]->visited = FALSE;
      distances[j] = -1;
      sigma[j] = 0;
    
    }
  
    src->rank = 0;
    //printf("src=%s\n", g->labels[src->vindex]);
    addheap(vheap, src);
    distances[src->vindex] = 0;
    sigma[src->vindex] = 1;
    //paths[src->vindex] = NULL;
  
    //printf("  pushing on stack:");

    while (v = (vertex*)removeheap(vheap))//remvertqueue(vQ))
    {
      if (!v->visited)
      {
        if (closeness)
          closeness[src->vindex] += distances[v->vindex];
        
        if (eccentricity && eccentricity[src->vindex] < distances[v->vindex])  
          eccentricity[src->vindex] = distances[v->vindex];
        
        pushlinkedlist(stack, v);
        //printf(" %s (d = %f)", g->labels[v->vindex], distances[v->vindex]);
        qsort(v->incidentedges, v->degree, sizeof(edge*), cmpedges);
      
        for (j = 0; j < v->degree; j++)
        {
          e = v->incidentedges[j];
      
          if (e->vertices[0] == v)  u = e->vertices[1];
      
          else if (e->vertices[1] == v)  u = e->vertices[0];
      
          else 
          {
            fprintf(stderr, "Error! Adjacency not found!\n");
            exit(1);
          }
        
          if (!u->visited)
          {
            newdist = distances[v->vindex];
            //if (ignoreweights)  newdist++;
            //else  
            newdist += e->weight;
      
            if (distances[u->vindex] < 0 || newdist < distances[u->vindex])
            {
              /* new, shorter path found, need to remove predecessors */
              //if (newdist < distances[u->vindex])
              //{
              while ((size = sizevector(predecessors[u->vindex])) > 0)  
                removevector(predecessors[u->vindex], size - 1) ;
              u->rank = newdist;
              addheap(vheap, u);
              distances[u->vindex] = newdist;
              //}
         
              addvector(predecessors[u->vindex], v);
              //distances[u->vindex] = newdist;
              sigma[u->vindex] = sigma[v->vindex];
              //paths[u->vindex] = v;
              //u->rank = newdist;
              //addheap(vheap, u);
            }
            
            else if (newdist == distances[u->vindex])
            {
              sigma[u->vindex] += sigma[v->vindex];
              addvector(predecessors[u->vindex], v);
            }
          }
        }
      
        v->visited = TRUE;
      }
    }
      
    for (j = 0; j < g->nvertices; j++)  delta[j] = 0;
      
    while (v = (vertex*)poplinkedlist(stack))
    { 
      //printf("\n  to: %s\n", g->labels[v->vindex]);
      //printf("  distance of v: %f\n", distances[v->vindex]);
      //printf("  sigma of v: %f\n", sigma[v->vindex]);
      //printf("  predecessors:");
      while ((size = sizevector(predecessors[v->vindex])) > 0)
      {
        u = (vertex*)removevector(predecessors[v->vindex],  size - 1);
        //printf(" %s", g->labels[u->vindex]);
        delta[u->vindex] += (sigma[u->vindex]/sigma[v->vindex]) * (1 + delta[v->vindex]);
      }
      if (v != src)  betweenness[v->vindex] += delta[v->vindex] / 2;
      //printf("\n  delta of v: %f\n", delta[v->vindex]);
      
    } 
    
    // empty predecessors lists
    for (j = 0; j < g->nvertices; j++)
    {
      while ((size = sizevector(predecessors[j])) > 0) removevector(predecessors[j], size);
    }
  } // end i loop over vertices
  
  
  for (i = 0; i < g->nvertices; i++)
  {
    if (closeness)
      closeness[i] = 1/closeness[i];
    
    if (eccentricity)
      eccentricity[i] = 1/eccentricity[i];
    
    freevector(predecessors[i]);
    free(predecessors[i]);
  }
  
  free(predecessors);
  freelinkedlist(stack);
  free(stack);
  free(distances);
  free(sigma);
  free(delta);
  freeheap(vheap);
  free(vheap);
  
  //return paths;
}


void printgraph(graph *g)
{
  int i;
  edge *e;
  vertex *u, *v;
  for (i = 0; i < g->nedges; i++)
  {
    e = g->edgelist[i];
    u = e->vertices[0];
    v = e->vertices[1];
    printf("%s:%s %f\n", g->labels[u->vindex], g->labels[v->vindex], e->weight);
  }
}

graph* testgraph()
{
  int nvertices = 7, nedges = 9;
  int i;
  vertex *u, *v;
  char *label;
  graph *g = (graph*)malloc(sizeof(graph));
  initgraph(g, nvertices, nedges);
  //g->labels = (char**)malloc(nvertices * sizeof(char*));

  for (i = 0; i < nvertices; i++)
  {
    label = (char*)malloc(2 * sizeof(char));
    label[0] = (char)(i + 49);
    label[1] = '\0';
    addvertexgraph(g, label);
    //g->labels[i] = (char*)malloc(1 * sizeof(char));
    //g->labels[i][0] = (char)(i + 49);
    //g->labels[i][1] = '\0';
  }
  
  u = g->vertices[0];
  v = g->vertices[1];
  addedgegraph(g, u, v, 1);
  
  v = g->vertices[2];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[1];
  v = g->vertices[2];
  addedgegraph(g, u, v, 1);
  
  v = g->vertices[4];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[2];
  v = g->vertices[3];
  addedgegraph(g, u, v, 1);
  
  v = g->vertices[4];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[3];
  v = g->vertices[5];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[4];
  v = g->vertices[6];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[5];
  v = g->vertices[6];
  addedgegraph(g, u, v, 1);
  
  
  return g;
}

graph* testgraph2()
{
  int nvertices = 6, nedges = 15;
  int i, j;
  int edgelengths[25] = { 2, 3, 2, 3, 3,  // from a
    -1, 1, 2, 3, 3,  // from b
    -1,-1, 2, 2, 1,  // from c
    -1,-1,-1, 1, 1,  // from d
    -1,-1,-1,-1, 2}; // from e
    vertex *u, *v;
    char *label;
    graph *g = (graph*)malloc(sizeof(graph));
    initgraph(g, nvertices, nedges);
    //g->labels = (char**)malloc(nvertices * sizeof(char*));

    for (i = 0; i < nvertices; i++)
    {
      label = (char*)malloc(2 * sizeof(char));
      label[0] = (char)(i + 97);
      label[1] = '\0';
      addvertexgraph(g, label);
      //g->labels[i] = (char*)malloc(1 * sizeof(char));
      //g->labels[i][0] = (char)(i + 97);
      //g->labels[i][1] = '\0';
    }
  
    for (i = 0; i < nvertices; i++)
    {
      u = g->vertices[i];
      for (j = i + 1; j < nvertices; j++)
      {
        v = g->vertices[j];
        addedgegraph(g, u, v, edgelengths[i * (nvertices-1) + (j - 1)]);
        printf("edge index: %d; edge length: %d\n", i * (nvertices-1) + (j - 1), edgelengths[i * (nvertices-1) + (j - 1)]);
      }
    }
  
  
    return g;
}

graph* testgraph3()
{
  int nvertices = 5, nedges = 5;
  int i;
  char *label;
  vertex *u, *v;
  graph *g = (graph*)malloc(sizeof(graph));
  
  initgraph(g, nvertices, nedges);
  
  //g->labels = (char**) malloc(nvertices * sizeof(char*));
  
  for (i = 0; i < nvertices; i++)
  {
    label = (char*)malloc(2 * sizeof(char));
    label[0] = (char)(i + 97);
    label[1] = '\0';
    addvertexgraph(g, label);
    //g->labels[i] = (char*)malloc(1 * sizeof(char));
    //g->labels[i][0] = (char)(i + 97);
    //g->labels[i][1] = '\0';
  }

  v = g->vertices[0];
  u = g->vertices[1];
  addedgegraph(g, u, v, 1);
  u = g->vertices[3];
  addedgegraph(g, u, v, 4); 
  
  v = g->vertices[1];
  u = g->vertices[2];
  addedgegraph(g, u, v, 1);
  
  v = g->vertices[3];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[4];
  addedgegraph(g, u, v, 1);
  
  return g;
}

/* fig. 4.4 in networks text */
graph* testgraph4()
{
  int nvertices = 10, nedges = 9;
  int i;
  char *label;
  vertex *u, *v;
  
  graph *g = (graph*)malloc(sizeof(graph));
  
  initgraph(g, nvertices, nedges);
  
  for (i = 0; i < nvertices; i++)
  {
    label = (char*)malloc(2 * sizeof(char));
    label[0] = (char)(i + 97);
    label[1] = '\0';
    printf("Label %d: %s\n", i, label);
    addvertexgraph(g, label);
  }
    
  u = g->vertices[0];
  v = g->vertices[1];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[2];
  addedgegraph(g, u, v, 1);
  
  v = g->vertices[3];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[4];
  addedgegraph(g, u, v, 1);
  
  v = g->vertices[5];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[6];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[7];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[8];
  addedgegraph(g, u, v, 1);
  
  u = g->vertices[9];
  addedgegraph(g, u, v, 1);
  
  return g;
}

