#include "cluster.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>

#define MINTAU 5
#define MINITER 1000
#define MAXITER 10000

int intcmpr(const void *p1, const void *p2)
{
  if (*(int*)p1 > *(int*)p2)  return 1;
  else if (*(int*)p1 < *(int*)p2)  return -1;
  
  return 0;
}

int dodostop(vector *tau, double alpha)
{
  int i;
  int k = sizevector(tau);
  double taustop, vhat = 0.;
  int taulast = *(int*)getvector(tau, k - 1);
  double numer = log(taulast - *(int*)getvector(tau, 0));
  //log(tau[k - 1] - tau[0]);
  
  for (i = 1; i < (k - 1); i++)
    vhat += numer - log(taulast - *(int*)getvector(tau, i));
    //log(tau[k - 1] - tau[i]);
  
  vhat /= k - 1;
  
  taustop = taulast + (taulast - *(int*)getvector(tau, 0)) / (pow(-log(alpha)/k, -vhat) - 1);
  //tau[k - 1] + (tau[k - 1] - tau[0]) / (pow(-log(alpha)/k, -vhat) - 1);
  
  return (int)(taustop + 0.5);
}

int* randomcentroids(int *centarray, int maxidx, int n)
{
  int nassigned = 0;
  int i;
  int cent;
  bool centfound;
  
  while (nassigned < n)
  {
    cent = rand() % maxidx;
    centfound = FALSE;
    for (i = 0; i < nassigned; i++)
      if (centarray[i] == cent)  centfound = TRUE;
    if (!centfound)
    {
      centarray[nassigned] = cent;
      nassigned++;
    }
  }
  
  /*printf("Initial centroids:");
  for (i = 0; i < n; i++)  printf(" %d", centarray[i]);
  printf("\n");*/
  
  return centarray;
}
int* locateclusters(int *clusters, int *centroids, double *distances, int ngenes, int k)
{
  int i, j;
  int cent;
  double dist;
  double score = 0;
  
  // loop over genes
  // loop over centroids
  // assign index of closest centroid to "clusters"
  //printf("clusters:");
  for (i = 0; i < ngenes; i++)
  {
    clusters[i] = 0;
    for (j = 1; j < k; j++)
    {
      dist = distances[i * ngenes + centroids[j]];
      
      /*****NEW: first condition *********/
      if (dist >= 0 && 
          (distances[i * ngenes + centroids[clusters[i]]] < 0 || 
          dist < distances[i * ngenes + centroids[clusters[i]]]))
        clusters[i] = j;
    }
    //printf(" %d", clusters[i]);
  }
  //printf("\n");
  return clusters;
}

//double chindex(int *clusters, double *distances, int ngenes, int k, double *Wk, double *Bk)

double chindex(int *clusters, double *distances, int ngenes, int k, double *Wk, double *Bk)
{
  int i, j, c;
  double ssbt = 0., sswt = 0.;
  double x2t, xt, xct, te;
  int nc, idx, ng;
  
  *Wk = 0.;
  *Bk = 0.;
  
  for (i = 0; i < ngenes; i++)
  //for (i = 0; i < cols; i++)
  {
    x2t = xt = ssbt = sswt = 0.;
    ng = 0;
    
    for (c = 0; c < k; c++)
    {
      xct = 0.;
      nc = 0;

      for (j = 0; j < ngenes; j++)
      //for (j = 0; j < rows; j++)     
      {
        if (clusters[j] == c)
        {
          idx = i * ngenes + j;
          //idx = j * cols + i;
          /**** NEW *****/
          if (distances[idx] >= 0)
          {
            x2t += distances[idx] * distances[idx];
            xct += distances[idx];
            //x2t += matrix[idx] * matrix[idx];
            //xct += matrix[idx];
            nc++;
            ng++;
          }
        }
      }
      
      //if (nc == 0)  printf("PROBLEM: nc for this cluster/column is 0!\n");
      
      if (nc)
      {
        xt += xct;
        te = (xct * xct)/nc;
        ssbt += te;
        sswt -= te;
      }
    }
    
    //if (ng != rows)  printf("Valid distances in column: %d genes: %d\n", ng, rows);
    ssbt -= (xt * xt)/ng;
    sswt += x2t;
    
    *Wk += sswt;
    *Bk += ssbt;
  }
  
  //fprintf(stderr, "Wk, Bk inside chindex is %f %f; ", *Wk, *Bk);
  return (*Bk / (k - 1)) / (*Wk / (ngenes - k));
}

int* locatecentroids(int *clusters, int *centroids, double *distances, int ngenes, int k)
{
  int i, j;
  double nodedist;
  double *clustdists = (double*)malloc(k * sizeof(double));
  int ndists;
  
  for (i = 0; i < k; i++)  clustdists[i] = -1;
  
  for (i = 0; i < ngenes; i++)
  {
    nodedist = 0;
    ndists = 0;
    for (j = 0; j < ngenes; j++)
    {
      /***** NEW ****/
      if (clusters[i] == clusters[j] && distances[i * ngenes + j] >= 0)
      {  
        nodedist += distances[i * ngenes + j]; 
        ndists++;
      }
    }
    
    if (ndists) // CAREFUL! Is it possible that a certain cluster will have NO valid distances?
    {
      nodedist /= ndists;
    
      if (clustdists[clusters[i]] < 0 || nodedist < clustdists[clusters[i]])
      {
        centroids[clusters[i]] = i;
        clustdists[clusters[i]] = nodedist;
      }
    }
  }
  
  for (i = 0; i < k; i++) 
  {
    if  (clustdists[i] < 0)
    {
      fprintf(stderr, "Error! All cluster distances seem to be invalid, no centroid!");
      exit(-1);
    }
  }
  
  //printf("clustdists:");
  //for (i = 0; i < k; i++)  printf(" %f", clustdists[i]);
  //printf("\n");
  
  free(clustdists);
  
  return clusters;
}

double iterkmeans(double *distances, int *clusters, int ngenes, int k)
{
  int i, j;
  int stopt95 = MINITER;
  int *tauitem;
  int delcent, newcent;
  int *newclusters = (int*)malloc(ngenes * sizeof(int));
  int *centroids, *newcentroids, *tmp;
  //FILE *scorefile = fopen("scores.txt", "w");
  double score, newscore;
  double sswk, ssbk;
  vector *tau = (vector*)malloc(sizeof(vector));
  initvector(tau);
        
  centroids = (int*)malloc(k * sizeof(int));
  newcentroids = (int*)malloc(ngenes * sizeof(int));
  
  randomcentroids(centroids, ngenes, k);
  qsort(centroids, k, sizeof(int), intcmpr);
  locateclusters(clusters, centroids, distances, ngenes, k);
  //score = kmeanscore(clusters, centroids, distances, ngenes, k);
  score = chindex(clusters, distances, ngenes, k, &sswk, &ssbk);
  score = sswk;
  tauitem = (int*)malloc(sizeof(int));
  *tauitem = 0;
  addvector(tau, tauitem);
  
  //i = 1;
  for (i = 1; (i < MINITER) || (i < stopt95) && (i < MAXITER); i++)
  {
    /*if (!(i % 10))
    {
      fprintf(scorefile, "%d %f\n", i, score);
      //printf("New score! %f\n", score);
    }*/
    delcent = rand() % k;
    while (TRUE)
    {
      newcent = rand() % ngenes;
      if (!(bsearch(&newcent, centroids, k, sizeof(int), intcmpr)))  break;
    }
    
    for (j = 0; j < k; j++)
    {
      if (j != delcent)  newcentroids[j] = centroids[j];
      else  newcentroids[j] = newcent;
      //printf(" %d", centroids[j]);
    }
    //printf("\n");
    
    qsort(newcentroids, k, sizeof(int), intcmpr);
    locateclusters(newclusters, newcentroids, distances, ngenes, k);
    //newscore = kmeanscore(newclusters, newcentroids, distances, ngenes, k);
    chindex(newclusters, distances, ngenes, k, &sswk, &ssbk);
    newscore = sswk;
    
    if (newscore < (score - SMALL))//keepnew(newscore, score))
    {
      //printf("New score! iter=%d, score=%f\n", i, newscore);
      tmp = newcentroids;
      newcentroids = centroids;
      centroids = tmp;
      
      tauitem = (int*)malloc(sizeof(int));
      *tauitem = i;
      addvector(tau, tauitem);
      if (sizevector(tau) >= MINTAU)
      {
        stopt95 = dodostop(tau, 0.05);
        //printf("Predicted number of iterations: %d\n", stopt95);
      }
      
      for (j = 0; j < ngenes; j++)  clusters[j] = newclusters[j];
      score = newscore;
    }
    
    //i++;
  }
  
  //printf("Finished %d iterations.... t95: %d\n", i, stopt95);
  
  //fclose(scorefile);
  
  freevector(tau);
  free(tau);
  free(newclusters);
  free(centroids);
  free(newcentroids);
  
  return score;
}

/* Note that this only does a single k-means, with one set of random 
 * centroids.
 * Also important: this k-means uses "median" centroids, and starts 
 * with distance data (NOT Euclidean coordinates). 
 */
double kmeans(double *distances, int *clusters, int ngenes, int k)
{
  int i, j;
  int *centroids = (int*)malloc(k * sizeof(int));
  double score, oldscore;
  double sswk, ssbk;

  randomcentroids(centroids, ngenes, k);
  qsort(centroids, k, sizeof(int), intcmpr);
  locateclusters(clusters, centroids, distances, ngenes, k);
  chindex(clusters, distances, ngenes, k, &sswk, &ssbk);
  score = sswk;

  oldscore = score + 1;
  
  while (oldscore > (score + SMALL))
  {
    oldscore = score;
    locatecentroids(clusters, centroids, distances, ngenes, k);
    qsort(centroids, k, sizeof(int), intcmpr);
    locateclusters(clusters, centroids, distances, ngenes, k);
    chindex(clusters, distances, ngenes, k, &sswk, &ssbk);
    score = sswk;
  }
  
  free(centroids);
  return score;//clusters;
} 

