#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "netcongruence.h"
#include "cluster.h"

//int PRINTCRAP = 0;
/*void emptybiparthash(hashtable *bptab)
{
  void *key;
  linkedlist *bplist = (linkedlist*)malloc(sizeof(linkedlist));
  initlinkedlist(bplist);
  
  resetiterhashtable(bptab);
  
  while ((key = nextkeyhashtable(bptab)) != NULL)
    enqueuelinkedlist(bplist, key);
  
  
  while((key = dequeuelinkedlist(bplist)) != NULL)
  {
    free(removehashtable(bptab, key));
    freetreebipart((treebipart*)key);
    free(key);
  }
  
  freelinkedlist(bplist);
  free(bplist);
} */

void initoptions(options *opts)
{
  opts->patchyrule = RELAXED;
  opts->clustmethod = EIGSPEC;
  opts->distmethod = BOTH;
  opts->hclmethod = COMPLETE;
  opts->cutoff = 0.5;
  opts->k = -1;
  opts->K = -1;
  
  opts->debug = FALSE;
  opts->verbose = FALSE;
  
  opts->taxafname = NULL;
  opts->trdistnfname = NULL;
  opts->burninfname = NULL;
  opts->minbl = -1;
}

void printoptions(options *opts)
{
  printf("patchyrule: %d\n", opts->patchyrule);
  printf("clustmethod: %d\n", opts->clustmethod);
  printf("distmethod: %d\n", opts->distmethod);
  printf("hclmethod:  %d\n", opts->hclmethod);
  printf("cutoff: %.2f\n", opts->cutoff);
  
  if (opts->taxafname)  printf("taxafname: %s\n", opts->taxafname);
  else  printf("(no taxafname)\n");
  
  if (opts->trdistnfname)  printf("trdistnfname: %s\n", opts->trdistnfname);
  else  printf("(no trdistnfname)\n");
  
  if (opts->burninfname)  printf("burninfname: %s\n", opts->burninfname);
  else printf("(no burninfname)\n");
  
  if (opts->debug)  printf("debug on, ");
  else  printf("debug off, ");
  
  if (opts->verbose)  printf("verbose on.\n");
  else  printf("verbose off.\n");
}

void setpatchyrule(options *opts, char *rulestr)
{
  int i;
  
  for (i = 0; i < strlen(rulestr); i++)
    rulestr[i] = tolower(rulestr[i]);
  
  if (strcmp(rulestr, "strict") == 0)  opts->patchyrule = STRICT;
  
  else if (strcmp(rulestr, "relaxed") == 0)  opts->patchyrule = RELAXED;
  
  else if (strcmp(rulestr, "oneprune") == 0)  opts->patchyrule = ONEPRUNE;
  
  else usage();
  
}

void setclustmethod(options *opts, char *clmstr)
{
  int i;
  
  for (i = 0; i < strlen(clmstr); i++)
    clmstr[i] = tolower(clmstr[i]);

  if (strcmp(clmstr, "kmeans") == 0)  opts->clustmethod = KMEANS;
  
  else if (strcmp(clmstr, "eigspec") == 0)  opts->clustmethod = EIGSPEC;
  
  else usage();
}

void setdistmethod(options *opts, char *distmstr)
{
  int i;
  
  for (i = 0; i < strlen(distmstr); i++)
    distmstr[i] = tolower(distmstr[i]);
  
  if (strcmp(distmstr, "ss") == 0)  opts->distmethod = SSDIST;
  
  else if (strcmp(distmstr, "ds") == 0)  opts->distmethod = DSDIST;
  
  else if (strcmp(distmstr, "both") == 0)  opts->distmethod = BOTH;
  
  else usage();
}

void sethclmethod(options *opts, char *hclstr)
{
  int i;
  
  for (i = 0; i < strlen(hclstr); i++)
    hclstr[i] = tolower(hclstr[i]);
  
  if (strcmp(hclstr, "complete") == 0)  opts->hclmethod = COMPLETE;
  
  else if ((strcmp(hclstr, "average") == 0) ||
           (strcmp(hclstr, "upgma") == 0))  
    opts->hclmethod = AVERAGE;
  
  else if (strcmp(hclstr, "single") == 0)  opts->hclmethod = SINGLE;
  
  
  else  usage();
}

/* Fill in missing filenames, if necessary */
void fixoptions(options *opts)
{
  if (! opts->taxafname)
  {
    opts->taxafname = (char*)malloc(11 * sizeof(char));
    strcpy(opts->taxafname, "nctaxa.txt");
    opts->taxafname[10] = '\0';
  }
  
  if (! opts->trdistnfname)
  {
    opts->trdistnfname = (char*)malloc(14 * sizeof(char));
    strcpy(opts->trdistnfname, "nctrdistn.txt");
    opts->trdistnfname[13] = '\0';
  }
  
  if (! opts->burninfname)
  {
    opts->burninfname = (char*)malloc(14 * sizeof(char));
    strcpy(opts->burninfname, "ncburnins.txt");
    opts->burninfname[13] = '\0';
  }
}

void freeoptions(options *opts)
{
  if (opts->taxafname)  free(opts->taxafname);
  if (opts->trdistnfname)  free(opts->trdistnfname);
  if (opts->burninfname)  free(opts->burninfname);
}

int gettdflist(char ***tdflist, char *filename, int maxfiles)
{
  FILE *tdf = fopen(filename, "r");
  
  if (! tdf)  return 0;
  
  char *linebuf = (char*)malloc(BUFSIZ * sizeof(char));
  int endchar, nfiles = 0;
  
  while (fgets(linebuf, BUFSIZ, tdf))
  {
    while (linebuf && isspace(linebuf[0]))  linebuf = &(linebuf[1]);
    
    endchar = strlen(linebuf) - 1;
    while (endchar >= 0 && isspace(linebuf[endchar]))  linebuf[endchar--] = '\0';
    
    if (endchar > 0)
    {
      if (nfiles >= maxfiles)
      {
        maxfiles *= 2;
        *tdflist = (char**)realloc(*tdflist, maxfiles * sizeof(char*));
      }
      
      (*tdflist)[nfiles] = (char*)malloc((endchar + 2) * sizeof(char));
      strcpy((*tdflist)[nfiles], linebuf);
          
      nfiles++;
    }
  }
  
  fclose(tdf);
  
  return nfiles;
}

void printbipartfreqs(hashtable *biparts)
{
  double freq;
  treebipart* tbp;
  int i;
  
  resetiterhashtable(biparts);
  while ((tbp = (treebipart*)nextkeyhashtable(biparts)) != NULL)
  {
    freq = *((double*)gethashtable(biparts, tbp));
    printf("%-12.6g ", freq);
    
    for (i = 0; i < tbp->nleaves; i++)
    {
      if (i && !(i % 10))  printf(" ");
      tbp->array[i] ? printf("*") : printf("-");
    }
    
    printf("\n");
  } 
}

int readtaxa(char ***taxaptr, char *filename, int maxtaxa)
{
  FILE *taxafile;
  char linebuf[BUFSIZ];
  int ntaxa = 0;
  int chidx;//, i;
  
  if (!(taxafile = fopen(filename, "r")))
  {
    fprintf(stderr, "Error, could not read taxa file %s!\n", filename);
    return -1;
  }
  
  while (fgets(linebuf, BUFSIZ, taxafile))
  {
    chidx = strlen(linebuf) - 1;
    
    while  (linebuf[chidx] == '\n' || linebuf[chidx] == ' ' || linebuf[chidx] == '\r')
    {
      linebuf[chidx] = '\0';
      chidx--;
    }
    
    if (ntaxa >= maxtaxa)
    {
      maxtaxa *= 2;
      *taxaptr = (char**)realloc(*taxaptr, maxtaxa * sizeof(char*));
    }
    
    (*taxaptr)[ntaxa] = (char*)malloc((1 + strlen(linebuf)) * sizeof(char));
    strcpy((*taxaptr)[ntaxa], linebuf);
    
    ntaxa++;
  }
  
  fclose(taxafile);
  
  return ntaxa;
}

int readburnins(int *burninptr, char *filename, int nburnins)
{
  FILE *burninfile;
  int nread = 0;
  char linebuf[BUFSIZ];
  
  if (!(burninfile = fopen(filename, "r")))
  {
    fprintf(stderr, "Error, could not read burnin file %s!", filename);
    return FALSE;
  }
  
  while (fgets(linebuf, BUFSIZ, burninfile))
  {
    burninptr[nread] = atoi(linebuf);
    nread++;
    if (nread > nburnins)  return FALSE;
  }
  
  fclose(burninfile);
  
  return (nburnins == nread);
}

/************************************
   DELETE THIS FUNCTION!

 ************************************/
/*int assignstringency(char *stringencystr)
{
  if (strcmp(stringencystr, "strict") == 0)  return STRICT;
  
  if (strcmp(stringencystr, "relaxed") == 0)  return RELAXED;
  
  if (strcmp(stringencystr, "oneprune") == 0)  return ONEPRUNE;
  
  return -1;
}*/

int probbiparts(hashtable *bipartprobs, char *filename, char **taxalist, int ntaxa, bool *genetaxa, int burnin, double minbl) 
{
  FILE *treefile;
  double *counter;
  const void *bpkey;
  linkedlist *removelist;
  int lineread = FALSE;
  int ntrees = 0;
  int maxlength = BUFSIZ;
  int nbufs = 1;
  char linebuf[BUFSIZ];
  char *buffer = (char*)malloc(BUFSIZ * sizeof(char));
  tree *t = NULL;
  int i, j, leafidx;
  int lastntaxa;
  int leafcount;
  
  buffer[0] = '\0';
  linebuf[0] = '\0';
    
  if (!(treefile = fopen(filename, "r")))
  {
    fprintf(stderr, "Error, could not read tree file %s!", filename);
    return FALSE;
  }
      
  while (fgets(linebuf, BUFSIZ, treefile))
  {
  
    if (nbufs * BUFSIZ > maxlength)
    {
      maxlength += BUFSIZ;
      buffer = (char*)realloc(buffer, maxlength * sizeof(char));
      buffer[strlen(buffer)] = '\0';
    }
    
    strcat(buffer, linebuf);
    
    if (linebuf[strlen(linebuf) - 2] == ';')  
    {
      lineread = TRUE;
    }
  
    if (lineread)
    {
      if (ntrees >= burnin)
      {

        //if (PRINTCRAP)  fprintf(stderr, "\t%d\n", ntrees);
        if (t)  
        {
          restring2tree(t, buffer);
          fixleavestree(t);
          if (minbl >= 0)  collapseshortbranchestree(t, minbl);
          
          if (ntrees == burnin || lastntaxa != t->nleaves)
            lastntaxa = t->nleaves;
        }

        else
        {
          t = (tree*)malloc(sizeof(tree));
          inittree(t, taxalist, ntaxa);
          string2tree(t, buffer);
          fixleavestree(t);
          
          if (minbl >= 0)  collapseshortbranchestree(t, minbl);
        }

        /*if (t->nleaves != 50)
	{
        fprintf(stderr, "Before counting biparts:");
	for (j = 0; j < (t->nleaves + t->ninternal); j++)
	  fprintf(stderr, " %d", t->nodearray[i]->binrep->nleaves);
	fprintf(stderr, "\n");
	}*/
	
        countbipartstree(t, bipartprobs);
      
        if (ntrees == burnin)
        {
          leafidx = 0;
          for (j = 0; j < ntaxa; j++)
          {
            if (leafidx < t->nleaves && strcmp(t->nodearray[leafidx]->label, taxalist[j]) == 0)
            {
              genetaxa[j] = TRUE;
              leafidx++;
            }
            
            else  genetaxa[j] = FALSE;
          }
        }
      }
      ntrees++;
      lineread = FALSE;

      strcpy(buffer, "");
      nbufs = 1;
    }
    
    else  nbufs ++;
    
  }
  
  fclose(treefile);
  freetree(t);
  free(t);
  free(buffer);
  
  if (ntrees < burnin)  
  {
    fprintf(stderr, "Error, burnin is greater than the number of trees in tree file %s!\n", filename);
    return FALSE;
  }
  
  if (ntrees < (burnin * 10))
    fprintf(stderr, "Warning: burnin is more than 10 %% of trees in tree file %s.\n", filename);
    
  ntrees -= burnin;
  
  resetiterhashtable(bipartprobs);
  //while (counter = (double*)nextvaluehashtable(bipartprobs))  *counter /= ntrees;
  
  removelist = (linkedlist*)malloc(sizeof(linkedlist));
  initlinkedlist(removelist);
  
  while (bpkey = nextkeyhashtable(bipartprobs))
  {
    counter = (double*)gethashtable(bipartprobs, bpkey);
    if ((*counter) < 2)      
      enqueuelinkedlist(removelist, bpkey);
      
    else
      *counter /= ntrees;
  }
  
  resetlinkedlist(removelist);
  
  while ((bpkey = dequeuelinkedlist(removelist)) != NULL)
  //{
    free(removehashtable(bipartprobs, bpkey));
    //freetreebipart((treebipart*)bpkey);
    //free(bpkey);
  //}
  
  freelinkedlist(removelist);
  free(removelist);
  
  return TRUE;
  
}

void revprunebiparts(hashtable *revprunedbiparts, hashtable *biparts, bool *genetaxa, int ntaxa)
{
  int ngenetaxa = 0;
  int *lookuptable = (int*)malloc(ntaxa * sizeof(int));
  treebipart *tbp, *comptbp;
  int bpsize;
  int i, j;
  const void *key;
  double *freqnew, *freqold, *freqoldcomp;
  
  j = 0;
  for (i = 0; i < ntaxa; i++)
  {
    if (genetaxa[i])  lookuptable[j++] = i;
  }
  
  ngenetaxa = j;
  
  resetiterhashtable(biparts);
  while (key = nextkeyhashtable(biparts))
  {
    tbp = (treebipart*)malloc(sizeof(treebipart));
    inittreebipart(tbp, ntaxa, -1);
    comptbp = (treebipart*)malloc(sizeof(treebipart));
    inittreebipart(comptbp, ntaxa, -1);
    
    if (ngenetaxa != ((treebipart*)key)->nleaves)
    {
      fprintf(stderr, "Error! Number of leaves counted differs from recorded number! (%d vs %d)\n", ngenetaxa, ((treebipart*)key)->nleaves);
      exit(1);
    }
    
    bpsize = 0;
    for (i = 0; i < ngenetaxa; i++)
    {
      if (((treebipart*)key)->array[i])
      {
        tbp->array[lookuptable[i]] = TRUE;
        bpsize++;
      }
      
      else  comptbp->array[lookuptable[i]] = TRUE;
    }
    
    if (bpsize >= (ngenetaxa - 1) || bpsize <= 1)
    {
      //fprintf(stderr, "Strange, found a trivial bipartition.\n");
      freetreebipart(tbp);
      freetreebipart(comptbp);
      free(tbp);
      free(comptbp);
    } 
    
    else
    {
      freqnew = (double*)gethashtable(biparts, key);
      freqold = (double*)gethashtable(revprunedbiparts, tbp);
      freqoldcomp = (double*)gethashtable(revprunedbiparts, comptbp);
      
      if (freqold || freqoldcomp)
      {
        fprintf(stderr, "Reverse-pruned to the same bipart!\n");
        fprintf(stderr, "tbp: (freq = %f)\n", *freqold);
        for (i = 0; i < ngenetaxa; i++)
        {
          if(tbp->array[i])
            fprintf(stderr, "*");
          else
            fprintf(stderr, "-");
        }
        fprintf(stderr, "\ncomptbp: (freq = %f)\n", *freqoldcomp);
        
        for (i = 0; i < ngenetaxa; i++)
        {
          if(comptbp->array[i])
            fprintf(stderr, "*");
          else
            fprintf(stderr, "-");
        }
        fprintf(stderr, "\n");

        
        exit(1);
      }
      
      else
      {
        freqold = (double*)malloc(sizeof(double));
        freqoldcomp = (double*)malloc(sizeof(double));
        *freqold = *freqoldcomp = *freqnew;
        inserthashtable(revprunedbiparts, tbp, freqold);
        inserthashtable(revprunedbiparts, comptbp, freqoldcomp);
      }
    }
  } 
   
  free(lookuptable);
}
double distbiparts(hashtable *biparts1, hashtable *biparts2)
{
  double dist = 0.0;
  double diffi;
  double *probptr;
  int i;
  const void *key;
  
  resetiterhashtable(biparts1);

  while (key = nextkeyhashtable(biparts1))
  {
    probptr = (double*)gethashtable(biparts2, key);
    
    if (probptr)
    {
      diffi = *((double*)gethashtable(biparts1, key));
      diffi = (*probptr) - diffi;
    }
    
    else
      diffi = *((double*)gethashtable(biparts1, key));
    
    dist += pow(diffi, 2); 
  }
  
  resetiterhashtable(biparts2);

  while(key = nextkeyhashtable(biparts2))
  {
    probptr = (double*)gethashtable(biparts1, key);
    
    if (!probptr)
    {
      diffi = *((double*)gethashtable(biparts2, key));
      dist += pow(diffi, 2);
    }
  }
  
  return sqrt(dist);
}

/**
 *  Return TRUE iff taxaA and taxaB have same presence/absence states
 * 
 *  Note that ntaxa is the number of taxa that are potentially present, i.e. 
 * present in the union of all genes compared
 */
bool havesameleaves(bool *taxaA, bool *taxaB, int ntaxa)
{
  int i;
  
  for (i = 0; i < ntaxa; i++)
    if (taxaA[i] != taxaB[i])  return FALSE;
  
  return TRUE;
}

bool* intersectTaxa(bool *intersection, bool *taxaA, bool *taxaB, int ntaxa)
{
  int i;
  
  for (i = 0; i < ntaxa; i++)
    intersection[i] = taxaA[i] && taxaB[i];
  
  return intersection;
}

/* prunedbiparts should already be initiated */
hashtable* prunebiparts(hashtable *prunedbiparts, bool *prunedtaxa, hashtable *biparts, bool *taxa, int ntaxa)
{
  int *lookup = (int*)malloc(ntaxa * sizeof(int));
  int i, j, k;
  int nprunedleaves, bpsize;
  treebipart *tbp;
  const void *key;
  double *freqnew, *freqold;
  
  j = k = 0;
  for (i = 0; i < ntaxa; i++)
  {
    if (taxa[i])
    {
      if (prunedtaxa[i])  
      {
        lookup[j] = k;
        k++;
      }
      
      else
      {
        lookup[j] = -1;
      }
      
      j++;
    }
  }
  
  nprunedleaves = k;
  
  resetiterhashtable(biparts);
  
  while (key = nextkeyhashtable(biparts))
  {
    bpsize = 0;
    tbp = (treebipart*)malloc(sizeof(treebipart));
    inittreebipart(tbp, nprunedleaves, -1);
    
    for (i = 0; i < ((treebipart*)key)->nleaves; i++)
    {
      if (lookup[i] > 0)
      {
        tbp->array[lookup[i]] = ((treebipart*)key)->array[i];
        if (tbp->array[lookup[i]])  bpsize++;
      }
    }
    
    /* Bipartition is trivial, ignore */
    if (bpsize >= tbp->nleaves - 1 || bpsize <= 1)  
    {
      freetreebipart(tbp);
      free(tbp);
    }

    else
    {
      /* complement bipart if necessary */
      if (tbp->array[0])
      {
        for (i = 0; i < tbp->nleaves; i++)
          tbp->array[i] = !(tbp->array[i]);
      }
      
      
      freqnew = (double*)gethashtable(biparts, key);
      freqold = (double*)gethashtable(prunedbiparts, tbp);
      
      if (freqold)
      {
        *freqold += *freqnew;
        freetreebipart(tbp);
        free(tbp);
      }
      
      else
      {
        freqold = (double*)malloc(sizeof(double));
        *freqold = *freqnew;
        inserthashtable(prunedbiparts, tbp, freqold);
      }
    }
  }
  
  free(lookup);
  
  return prunedbiparts;
}

int counttaxa(bool *taxapresent, int maxtaxa)
{
  int i, count = 0;
  
  for (i = 0; i < maxtaxa; i++)
    if (taxapresent[i])  count++;
  
  return count;
}

bool istoopatchy(hashtable *bipartsA, bool *taxaA, bool *taxaB, int maxtaxa, int stringency, double cutoff)
{
  // find set of taxa that are present in A but not B
  // get support value from bipartsA
  // if support < CUTOFF, return TRUE (or just print a message for now
  // don't forget to ignore trivial bipartitions
  // don't forget to complement bipartitiosn if 0th value is TRUE
  // if fewer than 4 taxa in common, return TRUE no matter what
  treebipart *missingB;
  int nleavesA = counttaxa(taxaA, maxtaxa); 
  int intersectsize = 0;
  int i;
  double *support;
  bool toopatchy = FALSE;
  
  missingB = (treebipart*)malloc(sizeof(treebipart));
  inittreebipart(missingB, nleavesA, -1);
  
  for (i = 0, nleavesA = 0; i < maxtaxa; i++)
  {
    if (taxaA[i])
    {
      if (taxaB[i])  
      { 
        missingB->array[nleavesA] = FALSE;
        intersectsize++;
      }
      
      else  missingB->array[nleavesA] = TRUE;
        
      nleavesA++;
    }
  }
  
  if  (intersectsize < 4)  toopatchy = TRUE;
  
  else if (stringency == RELAXED)  toopatchy = FALSE;

  else if (intersectsize < nleavesA - 1)
  {
    if (missingB->array[0])
    {
      for (i = 0; i < missingB->nleaves; i++)
        missingB->array[i] = !(missingB->array[i]);
    }
    
    support = (double*)gethashtable(bipartsA, missingB);
    
    if (!support)
      //printf("Support for bipartition containing missing taxa is 0.\n");
      // return TRUE;
      toopatchy = TRUE;
      
    else if (*support < cutoff)
      //printf("Low support for bipartition containing missing taxa: %g\n", *support);
      // return TRUE;
      toopatchy = TRUE;
    
    //else  printf("Support value looks OK: %g\n", *support);
      // return FALSE;
  }
  
  freetreebipart(missingB);
  free(missingB);
  
  return toopatchy;//FALSE;
}

double sumsquares(double *distances, int ngenes)
{
  double ss = 0.;
  double colsum, colsum2;
  int i, j;
  int validdists;
  
  //printf("in sum squares.\n");
  
  for (i = 0; i < ngenes; i++)
  {
    colsum = 0.;
    colsum2 = 0.;
    validdists = 0;
    for (j = 0; j < ngenes; j++)
    {
      if (distances[i * ngenes + j] >= 0) /****NEW****/
      {
        colsum += distances[i * ngenes + j];
        colsum2 += distances[i * ngenes + j] * distances[i * ngenes + j];
        validdists++;
      }
    }
    
    ss += colsum2 - (colsum * colsum / validdists);//ngenes);
  }
  //printf("Sum of squares: %f\n", ss);
  return ss;
}

double diff(double sswk, double sswklast, int k, int n)
{
  double diffk = pow(k - 1, 2./n) * sswklast;
  
  diffk -= pow(k, 2./n) * sswk;
  
  return diffk;
}

double klindex(double sswk, double sswknext, double sswklast, int k, int ngenes)
{ 
  double diffk = diff(sswk, sswklast, k, ngenes);
  double diffknext = diff(sswknext, sswk, k, ngenes);
  
  if (diffk < 0)  diffk *= -1;
  if (diffknext < 0)  diffknext *= -1;
  
  return diffk / diffknext;
}

int strcmpqsort(const void *p1, const void *p2)
{
  return strcmp(*(char* const*)p1, *(char* const*)p2);
}

void usage()
{
  fprintf(stderr, "Usage: netcongruence [options]\n\n");
  
  fprintf(stderr, "\t-p [rule]\tSet patchiness rule (relaxed, strict, or oneprune).\n");
  fprintf(stderr, "\t\t\tDEFAULT: relaxed\n\n");
  
  fprintf(stderr, "\t-c [method]\tSet clustering method (kmeans or eigspec).\n");
  fprintf(stderr, "\t\t\tDEFAULT: eigspec\n\n");
  
  fprintf(stderr, "\t-D [method]\tSet distance metric to use (ss, ds, or both)\n");
  fprintf(stderr, "\t\t\tss: Single-sided bipartition distances\n");
  fprintf(stderr, "\t\t\tds: Double-sided bipartition distances (no clustering will be performed)\n");
  fprintf(stderr, "\t\t\tDEFAULT: both\n\n");
  
  fprintf(stderr, "\t-H [method]\tSet hierarchical clustering method (complete, average, or single).\n");
  fprintf(stderr, "\t\t\tThis option applies only to eigen spectral clustering.\n");
  fprintf(stderr, "\t\t\tDEFAULT: complete\n\n");
  
  fprintf(stderr, "\t-x [cutoff]\tSpecify support value cutoff for pruning rule.\n");
  fprintf(stderr, "\t\t\tThis option applies only to the oneprune pruning rule.\n\n");
  fprintf(stderr, "\t\t\tDEFAULT: 0.05\n\n");
  
  fprintf(stderr, "\t-k [clusters]\tSpecify the number of clusters.\n");
  fprintf(stderr, "\t\t\tNote: should not be used in conjunction with -K\n");
  fprintf(stderr, "\t\t\tDEFAULT: estimate\n\n");
  
  fprintf(stderr, "\t-K [maxcluster]\tSpecify the maximum number of clusters.\n");
  fprintf(stderr, "\t\t\tNote: should not be used in conjunction with -k\n");
  fprintf(stderr, "\t\t\tDEFAULT: number of genes\n\n");
  
  fprintf(stderr, "\t-m [minbrlen]\t Specify minimum branch length.\n");
  fprintf(stderr, "\t\t\tThis option causes splits induced by branches shorter than minbrlen to be ignored\n");
  fprintf(stderr, "\t\t\tDEFAULT: never collapse branches\n\n");
  
  fprintf(stderr, "\t-t [taxafile]\tFilename for list of taxa.\n");
  fprintf(stderr, "\t\t\tDEFAULT: nctaxa.txt\n\n");
  
  fprintf(stderr, "\t-z [treefile]\tFilename for list of tree distribution files.\n");
  fprintf(stderr, "\t\t\tDEFAULT: nctrdistn.txt\n\n");
  
  fprintf(stderr, "\t-b [burninfile]\tFilename for list of burnins.\n");
  fprintf(stderr, "\t\t\tDEFAULT: ncburnins.txt\n\n");
  
  fprintf(stderr, "\t-h\t\tPrint this help message.\n\n");
  fprintf(stderr, "\t-d\t\tPrint debugging output.\n\n");
  fprintf(stderr, "\t-v\t\tPrint noisier output.\n\n");
  
  exit(0);
}

int main(int argc, char **argv)
{
  //struct dirent **dlist;
  int nfiles;
  int ntaxa;
  char **taxa = NULL, **dlist = NULL;
  bool *genetaxa, *prunedtaxa;
  int *ngenetaxa;
  int maxbiparts;
  int *burnins = NULL;
  hashtable **allbiparts = NULL;
  hashtable **revprunedbiparts = NULL;
  hashtable *prunedbipartsI, *prunedbipartsJ;
  double *distances = NULL, *dsdists = NULL, *ssdists = NULL, *dijkdists;
  double *eigvec = NULL;
  double *eigvals = NULL;
  vertex **dijkpaths;
  double score, chidx, kselectcrit, bestselectcrit;
  double *ssw, *ssb;
  int *clusters, *bestclusters;
  int bestk;
  int lowerk, upperk;
  char *treefilename;
  FILE *evfile = NULL;
  FILE *outfile = NULL;
  FILE *outfile2 = NULL;
  FILE *clustfile;
  graph *graphrep;
  double maxdist = 0;
  //int stringency;
  
  /* These should be CL options */
  //CLUSTMETHOD clustmethod = EIGSPEC;
  //HCLMETHOD hclmethod = COMPLETE;
  
  int i, j, k;
  int opt;
  
  options *ncopts = (options*)malloc(sizeof(options));
  initoptions(ncopts);
  
  while ((opt = getopt(argc, argv, "p:c:D:H:x:k:K:t:z:b:m:hdv")) != -1)
  {
    switch(opt)
    {
      case 'p':
        setpatchyrule(ncopts, optarg);
        break;
      
      case 'c':
        setclustmethod(ncopts, optarg);
        break;
      
      case 'D':
        setdistmethod(ncopts, optarg);
        break;
        
      case 'H':
        sethclmethod(ncopts, optarg);
        break;
        
      case 'x':
        ncopts->cutoff = atof(optarg);
        if ((ncopts->cutoff > 1) || (ncopts->cutoff < 0))
          usage();
        break;
        
      case 'k':
        if (ncopts->K != -1)  usage();
        ncopts->k = atoi(optarg);
        break;
        
      case 'K':
        if (ncopts->k != -1)  usage();
        ncopts->K = atoi(optarg);
        break;
      
      case 't':
        ncopts->taxafname = (char*)malloc((strlen(optarg) + 1) * sizeof(char));
        strcpy(ncopts->taxafname, optarg);
        ncopts->taxafname[strlen(optarg)] = '\0';
        break;
        
      case 'z':
        ncopts->trdistnfname = (char*)malloc((strlen(optarg) + 1) * sizeof(char));
        strcpy(ncopts->trdistnfname, optarg);
        ncopts->trdistnfname[strlen(optarg)] = '\0';
        break;
        
      case 'b':
        ncopts->burninfname = (char*)malloc((strlen(optarg) + 1) * sizeof(char));
        strcpy(ncopts->burninfname, optarg);
        ncopts->burninfname[strlen(optarg)] = '\0';
        break;
        
      case 'm':
        ncopts->minbl = atof(optarg);
        break;
      
      case 'd':
        ncopts->debug = TRUE;
        break;
        
      case 'v':
        ncopts->verbose = TRUE;
        break;
        
      case 'h':
      default:
        usage();
        break;
    }
  }
  
  fixoptions(ncopts);
  //printoptions(ncopts);
  
  if (optind != argc)  usage();
  
  srand(time(0));
  taxa = (char**)malloc(INITITEMS * sizeof(char*));
  ntaxa = readtaxa(&taxa, ncopts->taxafname, INITITEMS);
  if (ntaxa < 0)  goto cleanup;
  
  qsort(taxa, ntaxa, sizeof(char*), strcmpqsort);
  
  maxbiparts = 1 << ntaxa;
    
  dlist = (char**)malloc(INITMAXFILES * sizeof(char*));
  if (! (nfiles = gettdflist(&dlist, ncopts->trdistnfname, INITMAXFILES)))  goto cleanup;
  
  burnins = (int*)malloc(nfiles * sizeof(int));
  if(!readburnins(burnins, ncopts->burninfname, nfiles))  goto cleanup;
  
  allbiparts = (hashtable**)malloc(nfiles * sizeof(hashtable*));
  distances = (double*)malloc(nfiles * nfiles * sizeof(double));
  
  if (ncopts->distmethod != SSDIST)
  {
    revprunedbiparts = (hashtable**)malloc(nfiles * sizeof(hashtable*));
    dsdists = (double*)malloc(nfiles * nfiles * sizeof(double));
  }
  
  genetaxa = (bool*)malloc(nfiles * ntaxa * sizeof(bool));
  prunedtaxa = (bool*)malloc(ntaxa * sizeof(bool));
  ngenetaxa = (int*)malloc(nfiles * sizeof(int));

  printf("Reading trees and calculating distances...\n");
  for (i = 0; i < nfiles; i++)
  {
  
    //if (i == 60)  PRINTCRAP = 1;
    //else  PRINTCRAP = 0;
    distances[i * nfiles + i] = 0.0;
    
    allbiparts[i] = (hashtable*)malloc(sizeof(hashtable));
    inithashtable(allbiparts[i], hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
    
    
    probbiparts(allbiparts[i], dlist[i], taxa, ntaxa, &(genetaxa[ntaxa * i]), burnins[i], ncopts->minbl);
    
    ngenetaxa[i] = counttaxa(&(genetaxa[ntaxa * i]), ntaxa);
    
    if (ncopts->distmethod != SSDIST)
    {
      dsdists[i * nfiles + i] = 0.0;
      revprunedbiparts[i] = (hashtable*)malloc(sizeof(hashtable));
      inithashtable(revprunedbiparts[i], hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
      revprunebiparts(revprunedbiparts[i], allbiparts[i], &(genetaxa[ntaxa * i]), ntaxa);
    }
    
    for (j = 0; j < i; j++) 
    {
      
      if (ncopts->distmethod != SSDIST)
      {
        dsdists[i * nfiles + j] = distbiparts(revprunedbiparts[i], revprunedbiparts[j]) / sqrt(ngenetaxa[i] + ngenetaxa[j] - 6);
        dsdists[j * nfiles + i] = dsdists[i * nfiles + j];
      }
      
      if (havesameleaves(&(genetaxa[ntaxa * i]), &(genetaxa[ntaxa * j]), ntaxa))
      {
        distances[i * nfiles + j] = distbiparts(allbiparts[i], allbiparts[j]) / sqrt(2 * ngenetaxa[i] - 6);
        distances[j * nfiles + i] = distances[i * nfiles + j];
      }
      
      else if (ncopts->patchyrule == STRICT || 
	       (istoopatchy(allbiparts[i], &(genetaxa[ntaxa * i]), &(genetaxa[ntaxa * j]), ntaxa, ncopts->patchyrule, ncopts->cutoff) ||
               istoopatchy(allbiparts[j], &(genetaxa[ntaxa * j]), &(genetaxa[ntaxa * i]), ntaxa, ncopts->patchyrule, ncopts->cutoff)))
      {
        
        // Fix this, maybe, or just deal with it?
        distances[i * nfiles + j] = -1;
        distances[j * nfiles + i] = -1;
      }
      
      else
      {
        // get taxa intersection
        intersectTaxa(prunedtaxa, &(genetaxa[ntaxa * i]), &(genetaxa[ntaxa * j]), ntaxa);
        // prune biparts list
        // get distance
        prunedbipartsI = (hashtable*)malloc(sizeof(hashtable));
        inithashtable(prunedbipartsI, hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
        
        prunebiparts(prunedbipartsI, prunedtaxa, allbiparts[i], &(genetaxa[ntaxa * i]), ntaxa);
        
        prunedbipartsJ = (hashtable*)malloc(sizeof(hashtable));
        inithashtable(prunedbipartsJ, hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
        
        prunebiparts(prunedbipartsJ, prunedtaxa, allbiparts[j], &(genetaxa[ntaxa * j]), ntaxa);
        
        distances[i * nfiles + j] = distbiparts(prunedbipartsI, prunedbipartsJ)/sqrt(2 * counttaxa(prunedtaxa, ntaxa) - 6); 
            // /log(counttaxa(prunedtaxa, ntaxa)/3.0);
        distances[j * nfiles + i] = distances[i * nfiles + j];
        //fprintf(outfile, " %d", counttaxa(prunedtaxa, ntaxa));

        //emptybiparthash(prunedbipartsI);
        freehashtable(prunedbipartsI);
        free(prunedbipartsI);
        
        //emptybiparthash(prunedbipartsJ);
        freehashtable(prunedbipartsJ);
        free(prunedbipartsJ);
      }
      
      if (distances[i * nfiles + j] > maxdist)  maxdist = distances[i * nfiles + j];
      
     
      //exit(1);
    }
    
    // count biparts for all trees, skipping burnin
    // note that if burnin exceeds the number of lines in the file, GOTO cleanup and exit with an error
    //fprintf(outfile, "\n");  
    printf("  Finished %s.\n", dlist[i]); //dlist[i]->d_name);
    fflush(stdout);
  }
  
  //fclose(outfile);
  
  for (i = 0; i < nfiles; i++)  
  {
    //emptybiparthash(allbiparts[i]);
    if (allbiparts)
    {
      freehashtable(allbiparts[i]);
      free(allbiparts[i]);
    }
    //emptybiparthash(revprunedbiparts[i]);

    if (revprunedbiparts)
    {
      freehashtable(revprunedbiparts[i]);
      free(revprunedbiparts[i]);
    }
  }
  
  if (allbiparts)  free(allbiparts);
  if (revprunedbiparts)  free(revprunedbiparts);
  allbiparts = NULL;
  revprunedbiparts = NULL;
  
  free(genetaxa);
  free(prunedtaxa);
  
  graphrep = (graph*)malloc(sizeof(graph));
  buildgraph(graphrep, distances, dlist, nfiles, false);
  dijkdists = (double*)malloc(nfiles * sizeof(double));
  dijkpaths = (vertex**)malloc(nfiles * sizeof(vertex*));
  
  for (i = 0; i < nfiles; i++)
  {
    // Note: get rid of maxdist arg to Dijkstragraph!!!
    Dijkstragraph(graphrep, graphrep->vertices[i], dijkpaths, dijkdists, 100, FALSE);

    for (j = 0; j < nfiles; j++)  
    {
      if (distances[i * nfiles + j] < 0)  
        distances[i * nfiles + j] = dijkdists[j];
    }
  }  
  
  if (ncopts->distmethod == BOTH || ncopts->distmethod == SSDIST)
    outfile = fopen("ncbpdist.txt", "w");
  
  if (ncopts->distmethod == BOTH || ncopts->distmethod == DSDIST)
    outfile2 = fopen("ncdsdist.txt", "w");
    
  for (i = 0; i < nfiles; i++)
  {
    for (j = 0; j < i; j++)
    {
      if (ncopts->distmethod == BOTH || ncopts->distmethod == DSDIST)
      {
        fprintf(outfile2, "%s %s pp %.6f\n", dlist[i], dlist[j], dsdists[i * nfiles + j]);
      }
      
      if (ncopts->distmethod == BOTH || ncopts->distmethod == SSDIST)
      {
        if (distances[i * nfiles + j] > 0)
          fprintf(outfile, "%s %s pp %.6f\n", dlist[i], dlist[j], distances[i * nfiles + j]);
              
        else if (dijkdists[i * nfiles + j] > 0)
          fprintf(outfile, "%s %s pp %.6f\n", dlist[i], dlist[j], dijkdists[i * nfiles + j]);
        
        else 
        {
          fprintf(outfile, "%s %s pp %.6f\n", dlist[i], dlist[j], maxdist * 2);
          distances[i * nfiles + j] = maxdist * 2;
        }
      }
    }
  }
  
  if (outfile)
  {
    fclose(outfile);
    outfile = NULL;
  }
  
  if (outfile2)
  {
    fclose(outfile2);
    outfile2 = NULL;
  }
  
  freegraph(graphrep);
  free(graphrep);
  free(dijkdists);
  free(dijkpaths);
  
  for (i = 0; i < nfiles; i++)  free(dlist[i]);
  free(dlist);
  dlist = 0;
  
 
  // cluster distances
  clusters = (int*)malloc(nfiles * sizeof(int));
  bestclusters = (int*)malloc(nfiles * sizeof(int));
  
  ssw = (double*)malloc(nfiles * sizeof(double));
  ssb = (double*)malloc(nfiles * sizeof(double));
  ssw[0] = sumsquares(distances, nfiles);
  
  clustfile = fopen("ncclust.txt", "w");
  //fprintf(clustfile, " 2 %d\n", nfiles - 1);

  if ((ncopts->k > 1) && (ncopts->k < nfiles))
  {
    lowerk = ncopts->k;
    upperk = ncopts->k + 1;
  }
  
  else
  {
    if (ncopts->k > 0)  
    {
      fprintf(stderr, "Invalid number of clusters specified, will find optimal value.\n");
      ncopts->k = -1;
    }
    
    if (ncopts->clustmethod == EIGSPEC)
    {    
      eigvec = (double*)malloc(nfiles * nfiles * sizeof(double));
      if (! eigvec) 
      {
        fprintf(stderr, "Unable to allocate enough memory. Quitting.\n");
        exit(1);
      }
      
      eigvals = (double*)malloc(nfiles * sizeof(double));
      
      if (! eigvals)
      {
        fprintf(stderr, "Unable to allocate enough memory for eigenvalues. Quitting.\n");
        exit(1);
      }
        
      geteigvecs(distances, eigvals, eigvec, nfiles);

      lowerk = 2;//eigengap(eigvals, nfiles);
      upperk = nfiles;//lowerk + 1;
    }
    
    else
    {
      lowerk = 2;
      upperk = nfiles;
    }
    
    if (ncopts->K > 0)
    {
      if (ncopts->K > nfiles || ncopts->K < 2)
        fprintf(stderr, "Invalid maximum number of clusters specified, will use number of genes minus 1 (%d).\n", (nfiles - 1));
      else
        upperk = ncopts->K + 1;
    }
  }

  fprintf(clustfile, " %d %d\n", lowerk, upperk - 1);

  printf("\nClustering...\n");
  printf("     k      Score        SSW   CH_Index   KL_Index\n");

  if (ncopts->clustmethod == EIGSPEC)
  {
    eigvec = (double*)malloc(nfiles * nfiles * sizeof(double));
    eigvals = (double*)malloc(nfiles * sizeof(double));
    geteigvecs(distances, eigvals, eigvec, nfiles);
    
    /* comment this out */
    evfile = fopen("nceigvals.txt", "w");
    for (i = 0; i < nfiles; i++)  fprintf(evfile, " %.6g", eigvals[i]);
    
    fprintf(evfile, "\n");
    
    fclose(evfile);
    /* end region to comment out */
  }

  if (ncopts->distmethod == DSDIST)
  {
    ssdists = distances;
    distances = dsdists;
    dsdists = NULL;
  }

  for (k = lowerk; k < upperk ; k++)
  {
    /* use switch/case instead */
    if (ncopts->clustmethod == KMEANS)
      iterkmeans(distances, clusters, nfiles, k);

    
    else if (ncopts->clustmethod == EIGSPEC)
      score = esclust(eigvec, clusters, nfiles, k, ncopts->hclmethod);
    chidx = chindex(clusters, distances, nfiles, k, &(ssw[k - 1]), &(ssb[k - 1]));
    
    if (k > lowerk)  printf(" %10.4f\n", klindex(ssw[k - 2], ssw[k - 1], ssw[k - 3], k - 1, nfiles));
    
    printf("  %4d %10.4f %10.4f %10.4f", k, score, ssw[k - 1], chidx);
    
    if (ncopts->clustmethod == EIGSPEC)  kselectcrit = -score;
    else if (ncopts->clustmethod == KMEANS)  kselectcrit = chidx; 
    //printf("k = %d. Iterative k-means score: %f; CH index: %f; Wk: %f; Bk: %f\nClusters:", k, score, chidx, ssw[k-1], ssb[k-1]);
    for (i = 0; i < nfiles; i++)   fprintf(clustfile, " %d", clusters[i]); //printf("Wk: %f, Bk: %f", ssw[k-1], ssb[k-1]);
    fprintf(clustfile, "\n");    
    /*", k, score, chidx);*/
    
    if (k == lowerk || kselectcrit > bestselectcrit)
    {
      for (i = 0; i < nfiles; i++)  
      {
        bestclusters[i] = clusters[i];
        //printf(" %d", clusters[i]);
      }
      //printf("\n");
      bestselectcrit = kselectcrit;
      bestk = k;
      
    }
  }
  
  if (upperk == nfiles)
  {
    ssw[nfiles - 1] = 0.0;
    printf(" %10.4f\n", klindex(ssw[nfiles - 2], ssw[nfiles - 1], ssw[nfiles - 3], nfiles - 1, nfiles));  
  }
  
  else
  {
    if (ncopts->clustmethod == EIGSPEC)
      esclust(eigvec, clusters, nfiles, upperk, ncopts->hclmethod);
    else
      chindex(clusters, distances, nfiles, upperk, &(ssw[upperk - 1]), &(ssb[upperk - 1]));
    printf(" %10.4f\n", klindex(ssw[upperk - 2], ssw[upperk - 1], ssw[upperk - 3], upperk - 1, nfiles));
  }
  
  if (ncopts->clustmethod == EIGSPEC && ncopts->k < 0)
  {
    bestk = eigengap(eigvals, nfiles);
    
    if (ncopts->K > 0 && ncopts->K < bestk)
    {
      fprintf(stderr, "Optimal k exeeds specied kmax (%d vs. %d); using kmax\n", bestk, ncopts->K);
      bestk = ncopts->K;
    }
    
    esclust(eigvec, bestclusters, nfiles, bestk, ncopts->hclmethod);
    
  }
  
  if (ncopts->clustmethod == EIGSPEC)
  {
      /* save eigenvectors to a file */
    evfile = fopen("nceigvects.txt", "w");
    writeeigvects(evfile, eigvec, nfiles, bestk);
    fclose(evfile);
    
    if (ncopts->distmethod == BOTH)
    {
      geteigvecs(dsdists, eigvals, eigvec, nfiles);
      evfile = fopen("ncdseigvects.txt", "w");
      writeeigvects(evfile, eigvec, nfiles, bestk);
      
      fclose(evfile);
    }
  }
  
  printf("\nFINAL CLUSTERS (k = %d):", bestk);
  for (i = 0; i < nfiles; i++)  printf(" %d", bestclusters[i]);
  
  printf("\n");
  
  //clusters = {0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,2,2,2,2,2,3,3,3,3,3,3,3,3,3,
  //            4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
  //            7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,9,9,9,9,9}
  
  /*for (i = 0; i < 12; i++)  clusters[99 - i] = 0;
  for (i = 12; i < 20; i++)  clusters[99 - i] = 1;
  for (i = 20; i < 25; i++)  clusters[99 - i] = 2;
  for (i = 25; i < 34; i++)  clusters[99 - i] = 3;
  for (i = 34; i < 43; i++)  clusters[99 - i] = 4;
  for (i = 43; i < 52; i++)  clusters[99 - i] = 5;
  for (i = 52; i < 66; i++)  clusters[99 - i] = 6;
  for (i = 66; i < 75; i++)  clusters[99 - i] = 7;
  for (i = 75; i < 87; i++)  clusters[99 - i] = 8;
  for (i = 87; i < 100; i++)  clusters[99 - i] = 9;*/
  
  //chidx = chindex(clusters, distances, nfiles, 10, &(ssw[9]), &(ssb[9]));
  //printf("For real clusters, CH: %.2f; score: %.2f\n", chidx, ssw[9]);
  
  fclose(clustfile);
  
  free(ngenetaxa);
  
  if (eigvec)  free(eigvec);
  if (eigvals)  free(eigvals);
  
  /*for (i = 0; i < ntaxa; i++)  free(taxa[i]);
  free(taxa); */
  
  free(clusters);
  free(bestclusters);
  free(ssw);
  free(ssb);
  
  cleanup:
    freeoptions(ncopts);
    free(ncopts);

    if (taxa)
    {
      for (i = 0; i < ntaxa; i++)  free(taxa[i]);
      free(taxa);
    }
    
    if (dlist)
    {
      for (i = 0; i < nfiles; i++)  free(dlist[i]);
      free(dlist);
    }
    
    if (burnins)  free(burnins);
    
    if (allbiparts)
    {
      for (i = 0; i < nfiles; i++)  
      {
        //emptybiparthash(allbiparts[i]);
        freehashtable(allbiparts[i]);
        free(allbiparts[i]);
      }
      
      free(allbiparts);
    }
    
    if (revprunedbiparts)
    {
      for (i = 0; i < nfiles; i++)  
      {
        //emptybiparthash(revprunedbiparts[i]);
        freehashtable(revprunedbiparts[i]);
        free(revprunedbiparts[i]);
      }
      
      free(revprunedbiparts);
    }
    
    if (distances)  free(distances);
    if (dsdists)  free(dsdists);
    if (ssdists)  free(ssdists);

  return(0);
}

