#ifndef NETCONGRUENCE_H
#define NETCONGRUENCE_H

#include <unistd.h>
#include <stdbool.h>

#include "ds.h"
#include "cluster.h"

#define INITMAXFILES 10

typedef enum  
{
  STRICT = 10, 
  RELAXED,
  ONEPRUNE 
} PATCHYRULE;

typedef enum
{
  KMEANS = 20,
  EIGSPEC
} CLUSTMETHOD;

typedef enum
{
  SSDIST = 30,
  DSDIST,
  BOTH
} DISTMETHOD;

typedef struct 
{
  PATCHYRULE patchyrule;
  CLUSTMETHOD clustmethod;
  DISTMETHOD distmethod;
  HCLMETHOD hclmethod;
  
  double cutoff;
  double minbl;
  char *taxafname;
  char *trdistnfname;
  char *burninfname;
  int k;
  int K;
  
  bool debug;
  bool verbose;
  
} options;

/* Function prototypes */

void initoptions(options*);
void printoptions(options*);
void setpatchyrule(options*, char*);
void setclustmethod(options*, char*);
void setdistmethod(options*, char*);
void sethclmethod(options*, char*);
void fixoptions(options*);
void freeoptions(options*);
int gettdflist(char***, char*, int);
int readtaxa(char***, char*, int); 
int readburnins(int*, char*, int);
int probbiparts(hashtable*, char*, char**, int, bool*, int, double); 
double distbiparts(hashtable*, hashtable*);
void revprunebiparts(hashtable*, hashtable*, bool*, int);
void printbipartfreqs(hashtable*);
int strcmpqsort(const void*, const void*);
void usage();
double sumsquares(double*, int);
double diff(double, double, int, int);
double klindex(double, double, double, int, int);
bool havesameleaves(bool*, bool*, int);
bool* intersectTaxa(bool*, bool*, bool*, int);
hashtable* prunebiparts(hashtable*, bool*, hashtable*, bool*, int);
int counttaxa(bool*, int);
bool istoopatchy(hashtable*, bool*, bool*, int, int, double);


//void emptybiparthash(hashtable*);

#endif