import glob, os, sys, getopt, time, seqtools, threading, nctools

# place to look for conclustador
ncpath = '/home/jleigh/Development/nc/conclustador'
pbpath = '/usr/local/bin/pb'
stoppbpath = '/usr/local/bin/stoppb'



def runPB(genename, args, kwargs):

  pid = os.fork()
  if not pid:
    os.execvp(pbpath, (pbpath,) + tuple(kwargs['model'].split()) + ('-dc', '-d', '%s.seq' % genename, genename))

  f = None
  done = False
  kwargs['burnins'][genename] = -1

  while not done:
    if not f:
      try:
        f = open('%s.trace' % genename)
      except IOError:
        print 'no tracefile yet.'
        time.sleep(30)
        continue
    f.seek(0)

    try:
      burnin, gens = findburnin(f, col=2, countgen=True)
    except BurninError, be:
      if kwargs['debug']:
        sys.stderr.write('%s\n' % str(be))
      time.sleep(30)
      continue

    if burnin > 0:
      if kwargs['debug']:
        sys.stderr.write('%s burnt in! burnin=%d (%s generations)\n' % (genename, burnin, gens))
      kwargs['burnins'][genename] = burnin

    else:
      if kwargs['debug']:
        sys.stderr.write('not burnt in with ds %s\n' % genename)
      time.sleep(30)
      continue 

    if gens >= burnin * 11:
      os.system('%s %s' % (stoppbpath, genename))
      done = True
    else:
      time.sleep(30)

  try:
    os.wait()
  except OSError, ose:
    if kwargs['debug']:
      sys.stderr.write('got an OSError: %s (Ignoring)\n' % str(ose))
    pass

  f.close()
  
  print '  Finished %s.' % genename
  sys.stdout.flush()
  
def listsum(l, startidx, endidx):
  """
  Note that this function gracefully handles out-of-range indices
  """
  
  total = 0
  maxidx = min(len(l), endidx)
  minidx = max(0, startidx)
  
  for i in xrange(minidx, maxidx):
    total += l[i]
  
  return total
  
def liststdev(l, mean, startidx, endidx):
  total = 0
  maxidx = min(len(l), endidx)
  minidx = max(0, startidx)

  for i in xrange(minidx, maxidx):
    total += (mean - l[i]) ** 2
    
  return (total / (len(l) - 1)) ** 0.5  
  
def findburnin(tracefile, col=3, header=True, window=500, slide=50, neg=True, countgen=False):
  """Find burnin by comparing average lnL over "windows".
   
   The burnin is reached when log-likelihoods converge on a stable distribution.
   For this purpose, when the "plateau" region of the log-likelihood plot, that
   MCMC generation is the end of the burn-in. This function calculates the mean 
   log-likelihood over the final values of the MCMC (number of values defined by 
   the "window" parameter). The burnin is assumed to have been reached in the 
   beginning of the first window in which the mean log-likelihood falls within two
   standard deviations of that of the final window. Windows are overlapping, 
   moving a "slide" distance defined by the corresponding parameter.
   
   Parameters:
    tracefile: open file from which log-likelihoods should be read
    col: column in tracefile containing log-likelihoods
    header: if True, skip the first line in tracefile
    window: number of log-likelihoods to use when calculating local average
    slide: distance to move the window (must be <= window)
    neg: if True, assume that log-likelihood values are less than 0 (that is, 
         if False, values in tracefile are actually -lnL and therefore > 0)
       
   Return estimated generation where burnin ends, if found; -1 otherwise.
   """
  
  if slide > window:
    raise BurninError('window must be greater than slide distance!')
       
  
  if header:
    tracefile.readline()
    
  loglikes = []
  
  for line in tracefile:
    if not line:
      continue
    loglikes.append(float(line.split()[col]))
    
  ngen = len(loglikes)
  if ngen <= (window):
    raise BurninError('window must be at least the total number of mcmc samples!')
  
  finalavg = listsum(loglikes, ngen-window, ngen)/window
  finalstdev = liststdev(loglikes, finalavg, ngen - window, ngen)
  
  for i in xrange(0, ngen, slide):
    winavg = listsum(loglikes, i, i+window)/window
    #if winavg > finalavg:
    if winavg < (finalavg +  finalstdev) and winavg > (finalavg -  finalstdev):
      if countgen:
        return(max(i, window/2), ngen)
      return max(i, window/2)
  if countgen:  
    return (-1, ngen)
  return -1  
  
def runallPB(seqnames, model, ncpu, debug=False):
  
  print 'debug flag in runallPB:', debug 
  burnins = {}
  kwargs = {'model': model, 'burnins': burnins, 'debug': debug}

  lock = threading.Lock()

  threads = []
  for i in xrange(ncpu):
    t = nctools.PhyloThread(seqnames, runPB, lock, kwargs=kwargs)
    threads.append(t)
    if debug:
      sys.stderr.write('threads: %s\n' % str(threads))

  for t in threads:
    # hopefully make sure thread actually starts!
    lock.acquire()
    lock.release()
    t.start()

  if debug:
    sys.stderr.write('threads (2): %s\n' % str(threads))

  for t in threads:
    t.join()

  return burnins


def usage():
  sys.stderr.write('Usage: python ncbayes.py -m model [-c ncpu] [-h]\n\n')

  sys.stderr.write('  model: should be c20, cat, or wag\n')
  sys.stderr.write('  ncpu: number of CPUs to use (default = 1)\n\n')

  sys.exit(1)
  

class PhyloThread(threading.Thread):
  def __init__(self, genestack, phylofunc, lock, name=None, args=(), kwargs=()):
    threading.Thread.__init__(self, name=name)

    self.genestack = genestack
    self.phylofunc = phylofunc
    self.lock = lock
    self.args = args
    self.kwargs = kwargs
    
    if self.kwargs['debug']:
      sys.stderr.write('%s\n' % self.getName())

  def run(self):
    if self.kwargs['debug']:
      sys.stderr.write('Starting run with thread %s\n' % self.getName())
    while True:
      self.lock.acquire()
      if len(self.genestack) == 0:
        self.lock.release()
        break
      genename = self.genestack.pop()
      if self.kwargs['debug']:
        sys.stderr.write('popped gene %s with thread %s\n' % (genename, self.getName()))
        sys.stderr.write('genes remaining in queue: %s\n' % str(self.genestack))
      self.lock.release()
      if self.kwargs['debug']:
        sys.stderr.write('About to enter phylofunc %s with gene %s\n' % (self.phylofunc.__name__, genename))
      self.phylofunc(genename, self.args, self.kwargs)
    if self.kwargs['debug']:
      sys.stderr.write('Done thread %s\n' % self.getName())

class BurninError(Exception):
  def __init__(self, value=''):
    self.value = value
  def __str__(self):
    return repr(self.value)
    

if __name__ == '__main__':
  # arguments:
  # -m model
  # -c ncpu (default 0)
  # -h
  
  
  try:
    optdict, args = nctools.dogetopt(sys.argv[1:], 'm', 'm:c:dh')
  except getopt.GetoptError:
    usage()


  if 'h' in optdict:
    usage()

  if 'd' in optdict:
    debug = True
  else:
    debug = False

  if 'c' in optdict:
    try:
      ncpu = int(optdict['c'])
    except ValueError:
      usage()
  else:
    ncpu = 1
    
  if optdict['m'] == 'c20':
    model = '-catfix c20'
  elif optdict['m'] == 'cat':
    model = '-cat'
  elif optdict['m'] == 'wag':
    model = '-wag'

  seqnames = [f.replace('.seq', '') for f in glob.glob('*.seq')]
  seqnames.sort()
  seqnamescopy = seqnames[:]
  seqnames.reverse()
  
  
  burnindict = runallPB(seqnames, model, ncpu, debug)
  tracelist = glob.glob('*.treelist')
  
  seqnames = seqnamescopy
  
  if len(tracelist) != len(seqnames):
    sys.stderr.write('seqnames: %s\n' % str(seqnames))
    sys.stderr.write('tracelist: %s\n' % str(tracelist))
    sys.stderr.write('An error occurred with PhyloBayes: not enough tree distributions found!\n')
    sys.exit(1)

  burninfile = open('ncburnins.txt', 'w')
  trdistfile = open('nctrdistn.txt', 'w')
  
  for align in seqnames:
    burninfile.write('%d\n' % burnindict[align])
    trdistfile.write('%s.treelist\n' % align)
  
  burninfile.close()
  trdistfile.close()
  
  taxa = nctools.countaligntaxa(seqnames).keys()
  taxa.sort()
  
  taxafile = open('nctaxa.txt', 'w')
  for t in taxa:
    taxafile.write('%s\n' % t)
  taxafile.close()
  
  os.system(ncpath)
  
