import glob, os, sys, getopt, seqtools, nctools, threading

puzzlepath = '/usr/bin/puzzle'
bionjpath = '/usr/local/bin/bionj'
ncpath = '/home/jleigh/Development/nc/trunk/conclustador'

puzzlemodels = ('auto', 'dayhoff', 'jtt', 'mtrev24', 'blosum62', 'vt', 'wag')

def runallpuzzle(seqnames, nboots, model, ncpu, debug=False):
  
  kwargs = {'nboots':nboots, 'model':model, 'debug': debug}
  aligns = seqtools.getaligns(seqnames)
  seqnames = [f.replace('.seq', '') for f in seqnames]
  aligndict = {}
  
  taxadict = {}
  treef = open('nctrdistn.txt', 'w')
  burnf = open('ncburnins.txt', 'w')
  
  for i in xrange(len(aligns)):
    a = aligns[i]
    truncatenames(a, taxadict)
    aligndict[seqnames[i]] = a
    treef.write('%s.boottrees\n' % seqnames[i])
    burnf.write('0\n') 
  treef.close()
  burnf.close()

  taxaf = open('nctaxa.txt', 'w')
  for name in taxadict.values():
    taxaf.write('%s\n' % name)
  taxaf.close()
        
  
  kwargs['aligns'] = aligndict
  seqnames.reverse()
  
  lock = threading.Lock()
  
  threads = []
  
  for i in xrange(ncpu):
    t = nctools.PhyloThread(seqnames, runpuzzle, lock, kwargs=kwargs)
    threads.append(t)
    
  if debug:
    sys.stderr.write('threads: %s\n' % str(threads))
    
  for t in threads:
    lock.acquire()
    lock.release()
    t.start()
    
  for t in threads:
    t.join()  

def truncatenames(alignment, taxadict):
  for seq in alignment:
    if seq.name in taxadict:
      seq.name = taxadict[seq.name]
    #elif len(seq.name) <= 10:
    #  taxadict[seq.name] = seq.name
    else:
      if len(seq.name) > 10:
        print 'seqname too long, will truncate:', seq.name
      shortname = seq.name[:10]
      allshortnames = set(taxadict.values())
      namecount = 2
      
      while shortname in allshortnames:
        print shortname, 'still in allshortnames, renaming (longname: %s)' % seq.name
        countstr = str(namecount)
        shortname = seq.name[:(10 - len(countstr))] + countstr
        namecount += 1
                
      taxadict[seq.name] = shortname
      seq.name = shortname
      
def runpuzzle(genename, args, kwargs):
  #in kwargs, nboots, model, debug
  
  align = kwargs['aligns'][genename]
  
  try:
    modelidx = puzzlemodels.index(kwargs['model'])
  except ValueError:
    sys.stderr.write('Model "%s" not found among puzzle-supported models. Using "auto" model.\n')
    modelidx = 0
  
  bootmatfile = open('%s.bootdists' % genename , 'w')
  
  
  for b in xrange(nboots):
    bootname = '%s.boot_%d.seq' % (genename, b)
    seqtools.writephy(seqtools.bootstrap(align), bootname, relaxed=False)
    
    inputstr = 'k\nk\nk\ne\n'
    
    for i in xrange(modelidx):
      inputstr += 'm\n'
    
    inputstr += 'y\n'

    nctools.runpipedbin((puzzlepath, bootname), inputstr, kwargs['debug'])
    
    try:
      distfile = open('%s.dist' % bootname)
      bootmatfile.write(distfile.read())
      distfile.close()
      bootmatfile.flush()
      
      os.remove('%s.puzzle' % bootname)
      os.remove('%s.dist' % bootname)
      os.remove('%s' % bootname)
    except IOError:
      sys.stderr.write('Error opening file %s.dist!\n' % bootname)
      sys.exit(1)
      
    except OSError:
      sys.stderr.write('Error deleting puzzle output files %s.dist and %s.puzzle!\n' % (bootname, bootname))
      sys.exit(1)
  
  bootmatfile.close()
  
  inputstr = '%s.bootdists\n%s.boottrees\n' % (genename, genename)
  
  nctools.runpipedbin(bionjpath, inputstr)
    
  sys.stdout.write('  Finished %s.\n' % genename)
  sys.stdout.flush()
  
def usage():
  
  sys.stderr.write('Usage: python ncdist.py -m model -n reps [-c ncpu] [-h]\n\n')
  
  sys.stderr.write('  model: should be dayhoff, jtt, mtrev24, blosum62, vt, or wag\n')
  sys.stderr.write('  reps: number of bootstrap replicates to perform for each alignment\n') 
  sys.stderr.write('  ncpu: number of CPUs to use (default = 1)\n\n')
  
  sys.exit(1)

if __name__ == '__main__':
  
  # should be user-specified, along with conclustador command
  #nboots = 1000
  #ncpu = 8
  #model = 'wag'
  #debug = False
  
  try:
    optdict, args = nctools.dogetopt(sys.argv[1:], 'mn', 'm:n:c:dh')
  except getopt.GetoptError:
    usage()
    
  if 'h' in optdict:
    usage()
    
  if 'd' in optdict:
    debug = True
  else:
    debug = False
    
  if 'c' in optdict:
    try:
      ncpu = int(optdict['c'])
    except ValueError:
      usage()
  else:
    ncpu = 1
    
  
  if optdict['m'] in set(puzzlemodels):
    model = optdict['m']
  else:
    sys.stderr.write('Invalid model: %s\n' % optdict['m'])
    usage()
    
  model = optdict['m']
  
  try:
    nboots = int(optdict['n'])
  except ValueError:
    usage()
  
  seqnames = glob.glob('*.seq')
  seqnames.sort()
  
  runallpuzzle(seqnames, nboots, model, ncpu, debug=debug)
  
  os.system(ncpath)
