import nctools, threading, glob, sys, os, getopt, seqtools, random

raxmlpath = '/usr/local/bin/raxmlHPC'
ncpath = '/home/jleigh/Development/nc/trunk/conclustador'

raxmlmodels = set(('dayhoff', 'dcmut', 'jtt', 'mtrev', 'wag', 'rtrev', 'cprev', 'vt', 'blosum62', 'mtmam', 'gtr'))

def runallraxml(seqnames, nboots, model, ncpu, slowboot=False, debug=False):

  kwargs = {'nboots': nboots, 'model': model, 'slowboot': slowboot, 'debug': debug}
  
  jobs = [{'name': seq, 'seed': random.randint(0, 1E6)} for seq in seqnames]
  
  lock = threading.Lock()
  
  threads = []
  
  for i in xrange(ncpu):
    t = nctools.PhyloThread(jobs, runraxml, lock, kwargs=kwargs)
    threads.append(t)
    
  if debug:
    sys.stderr.write('threads: %s\n' % str(threads))
  
  for t in threads:
    lock.acquire()
    lock.release()
    t.start()
  
  for t in threads:
    t.join()


def runraxml(genedict, args, kwargs):

  seed = genedict['seed']
  seedstr = str(seed)
  
  bootstr = str(kwargs['nboots'])
  
  genename = genedict['name']
  
  
  if kwargs['debug']:
    sys.stderr.write('Using random seed: %d\n' % seed)
  
  
  # raxmlpath -m %s -s %s.seq -n %s -f b -n %d
  if kwargs['slowboot']:
    model = 'PROTGAMMA' + kwargs['model'].upper()
    nctools.runpipedbin((raxmlpath, '-s', '%s.seq' % genename, '-n', genename, '-m', model, '-N', bootstr, '-b', seedstr), debug=kwargs['debug'])
    
  else:
    model = 'PROTCAT' + kwargs['model'].upper()
    nctools.runpipedbin((raxmlpath, '-s', '%s.seq' % genename, '-n', genename, '-m', model, '-N', bootstr, '-x', seedstr), debug=kwargs['debug'])
    
  os.remove('RAxML_info.%s' % genename)
    
  sys.stdout.write('  Finished %s.\n' % genename)
  sys.stdout.flush()

def usage():
  
  sys.stderr.write('Usage: python ncml.py -m model -n reps [-c ncpu] [-s] [-h]\n\n')
  
  sys.stderr.write('  model: should be dayhoff, dcmut, jtt, mtrev, wag, rtrev, cprev, vt, blosum62, mtmam, gtr\n')
  sys.stderr.write('  reps: number of bootstrap replicates to perform for each alignment\n') 
  sys.stderr.write('  ncpu: number of CPUs to use (default = 1)\n')
  sys.stderr.write('  -s: specifies slower bootstrapping algorithm.\n\n')
  
  sys.exit(1)

if __name__ == '__main__':
  
  try:
    optdict, args = nctools.dogetopt(sys.argv[1:], 'mn', 'm:n:c:sdh')
  except getopt.GetoptError:
    usage()
    
  if 'h' in optdict:
    usage()
    
  if 'd' in optdict:
    debug = True
  else:
    debug = False
    
  if 's' in optdict:
    slow = True
  else:
    slow = False
    
  if 'c' in optdict:
    try:
      ncpu = int(optdict['c'])
    except ValueError:
      usage()
  else:
    ncpu = 1
    
  

  if (optdict['m'] in raxmlmodels) or \
    ((optdict['m'][-1] == 'f') and (optdict['m'][:-1] in raxmlmodels)):
    model = optdict['m']
    
  else:
    sys.stderr.write('Invalid model: %s\n' % optdict['m'])
    usage()
  
  try:
    nboots = int(optdict['n'])
  except ValueError:
    usage()
  
  
  seqnames = [f.replace('.seq', '') for f in glob.glob('*.seq')]
  seqnames.sort()
  
  trdistnfile = open('nctrdistn.txt', 'w')
  for s in seqnames:
    trdistnfile.write('RAxML_bootstrap.%s\n' % s)
  trdistnfile.close()
  
  seqnames.reverse()
  
  runallraxml(seqnames, nboots, model, ncpu, slow, debug)

  taxa = nctools.countaligntaxa(seqnames).keys()
  taxa.sort()
  
  taxafile = open('nctaxa.txt', 'w')
  for t in taxa:
    taxafile.write('%s\n' % t)
  taxafile.close()
  
  burninfile = open('ncburnins.txt', 'w')
  for b in xrange(len(seqnames)):
    burninfile.write('0\n')
  burninfile.close()
  
  os.system(ncpath)
