import threading, getopt, glob, sys, subprocess

def runpipedbin(args, inputstr='', debug=False):
  if debug:
    sys.stderr.write('About to run process %s\n' % str(args))
    sys.stderr.write('Using input string: %s\n' % inputstr)
  
  if inputstr:
    proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    
  else:
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  
  data = proc.communicate(inputstr)
  
  if debug:
    if data[1]:
      sys.stderr.write('Stderr (should not see this!):\n')
      sys.stderr.write(data[1])
    if data[0]:
      sys.stderr.write(data[0])
      

def dogetopt(argv, reqdopts, legalopts):
  """Return a dictionary of options, check to make sure reqdopts are there."""

  opts, args = getopt.getopt(argv, legalopts)

  optdict = {}

  for (opt, val) in opts:
    optdict[opt[1:]] = val

  for opt in reqdopts:
    if opt not in optdict:
      raise getopt.GetoptError('Missing required option!')

  return (optdict, args)
  
def countaligntaxa(alignfiles):
  
  taxacount = {}
  
  for fname in alignfiles:
    af = open('%s.seq' % fname)
    
    header = af.readline()
    try:
      wordlist = header.split()
      nchar = int(wordlist[0])
      ntaxa = int(wordlist[1])
      
    except (ValueError, IndexError):
      af.close()
      sys.stderr.write('Error! Input file %s.seq appears not to be PHYLIP format!\n' % fname)
    
    linecount = 0
    for line in af:
      wordlist = line.split()
      
      if len(wordlist) == 2:
        taxname = wordlist[0]
      else:
        taxname = line[:10].strip()
        
      try:
        taxacount[taxname] += 1
      except KeyError:
        taxacount[taxname] = 1
      
      linecount += 1
      if linecount >= ntaxa:
        break
      
    af.close()
      
  return taxacount


class PhyloThread(threading.Thread):
  def __init__(self, genestack, phylofunc, lock, name=None, args=(), kwargs=()):
    threading.Thread.__init__(self, name=name)

    self.genestack = genestack
    self.phylofunc = phylofunc
    self.lock = lock
    self.args = args
    self.kwargs = kwargs
    
    if self.kwargs['debug']:
      sys.stderr.write('%s\n' % self.getName())

  def run(self):
    if self.kwargs['debug']:
      sys.stderr.write('Starting run with thread %s\n' % self.getName())
    while True:
      if self.kwargs['debug']:
        sys.stderr.write('gene stack before popping: %s\n' % str(self.genestack))
      self.lock.acquire()

      if len(self.genestack) == 0:
        self.lock.release()
        break
      genename = self.genestack.pop()
      runset = set(glob.glob('*.run'))
      if '%s.run' % genename in runset:
        if self.kwargs['debug']:
          sys.stderr.write('Already ran PB with gene %s, returning.\n' % genename)
        self.lock.release()
        continue
      if self.kwargs['debug']:
        sys.stderr.write('popped gene %s with thread %s\n' % (str(genename), self.getName()))
        sys.stderr.write('genes remaining in queue: %s\n' % str(self.genestack))
      self.lock.release()
      if self.kwargs['debug']:
        sys.stderr.write('About to enter phylofunc %s with gene %s\n' % (self.phylofunc.__name__, str(genename)))
      self.phylofunc(genename, self.args, self.kwargs)
    if self.kwargs['debug']:
      sys.stderr.write('Done thread %s\n' % self.getName())

