#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "netcongruence.h"
#include "ds.h"

void readtrees(hashtable *bpp, char *filename, char **taxalist, int ntaxa)
{
  const int burnin = 250;
  FILE *treefile = NULL;
  int maxlength = BUFSIZ;
  int nbufs;
  int ntrees;
  int treecount = 0;
  char linebuf[BUFSIZ];
  char *buffer = (char*)malloc(BUFSIZ * sizeof(char));
  tree *t = NULL;
  bool lineread = false;
  
  buffer[0] = '\0';
  linebuf[0] = '\0';
  
  if (!(treefile = fopen(filename, "r")))
  {
    fprintf(stderr, "Could not read treefile!\n\n");
    exit(1);
  }
  
  nbufs = 1;
  while (fgets(linebuf, BUFSIZ, treefile))
  {
    //printf("Got line: %s\n", linebuf);
    if (nbufs * BUFSIZ > maxlength)
    {
      maxlength += BUFSIZ;
      buffer = (char*)realloc(buffer, maxlength * sizeof(char));
      buffer[strlen(buffer)] = '\0';
    }
    
    strcat(buffer, linebuf);
    
    if (linebuf[strlen(linebuf) - 2] == ';')
      lineread = true;
    
    if (lineread)
    {
      if (treecount >= burnin)
      {
        if (t)  
        {
          restring2tree(t, buffer);
          fixleavestree(t);
        }
      
        else
        {
          t = (tree*)malloc(sizeof(tree));
          inittree(t, taxalist, ntaxa);
          string2tree(t, buffer);
          fixleavestree(t);
        }
        
        countbipartstree(t, bpp);

      }
      
      treecount++;
      nbufs = 1;
      buffer[0] = '\0';
    }
    
    else  nbufs++;
      
  }
  
  if (t)
  {
    freetree(t);
    free(t);
  }
  
  free(buffer);
  if (treefile)  fclose(treefile);
}

int probbiparts(hashtable *bipartprobs, char *filename, char **taxalist, int ntaxa, bool *genetaxa, int burnin) 
{
  FILE *treefile;
  double *counter;
  const void *bpkey;
  linkedlist *removelist;
  int lineread = FALSE;
  int ntrees = 0;
  int maxlength = BUFSIZ;
  int nbufs = 1;
  char linebuf[BUFSIZ];
  char *buffer = (char*)malloc(BUFSIZ * sizeof(char));
  tree *t = NULL;
  int i, j, leafidx;
  int lastntaxa;
  int leafcount;
  
  buffer[0] = '\0';
  linebuf[0] = '\0';
    
  if (!(treefile = fopen(filename, "r")))
  {
    fprintf(stderr, "Error, could not read tree file %s!", filename);
    return FALSE;
  }
      
  while (fgets(linebuf, BUFSIZ, treefile))
  {
  
    if (nbufs * BUFSIZ > maxlength)
    {
      maxlength += BUFSIZ;
      buffer = (char*)realloc(buffer, maxlength * sizeof(char));
      buffer[strlen(buffer)] = '\0';
    }
    
    strcat(buffer, linebuf);
    
    if (linebuf[strlen(linebuf) - 2] == ';')  
    {
      lineread = TRUE;
    }
  
    if (lineread)
    {
      if (ntrees >= burnin)
      {

        //if (PRINTCRAP)  fprintf(stderr, "\t%d\n", ntrees);
        if (t)  
        {
          restring2tree(t, buffer);
          fixleavestree(t);
          
          if (ntrees == burnin || lastntaxa != t->nleaves)
            lastntaxa = t->nleaves;
        }

        else
        {
          t = (tree*)malloc(sizeof(tree));
          inittree(t, taxalist, ntaxa);
          string2tree(t, buffer);
          fixleavestree(t);
        }

        countbipartstree(t, bipartprobs);
      
        if (ntrees == burnin)
        {
          leafidx = 0;
          for (j = 0; j < ntaxa; j++)
          {
            if (leafidx < t->nleaves && strcmp(t->nodearray[leafidx]->label, taxalist[j]) == 0)
            {
              genetaxa[j] = TRUE;
              leafidx++;
            }
            
            else  genetaxa[j] = FALSE;
          }
        }
      }
      ntrees++;
      lineread = FALSE;

      strcpy(buffer, "");
      nbufs = 1;
    }
    
    else  nbufs ++;
    
  }
  
  fclose(treefile);
  freetree(t);
  free(t);
  free(buffer);
  
  if (ntrees < burnin)  
  {
    fprintf(stderr, "Error, burnin is greater than the number of trees in tree file %s!\n", filename);
    return FALSE;
  }
  
  if (ntrees < (burnin * 10))
    fprintf(stderr, "Warning: burnin is more than 10 %% of trees in tree file %s.\n", filename);
    
  ntrees -= burnin;
  
  resetiterhashtable(bipartprobs);
  //while (counter = (double*)nextvaluehashtable(bipartprobs))  *counter /= ntrees;
  
  removelist = (linkedlist*)malloc(sizeof(linkedlist));
  initlinkedlist(removelist);
  
  while (bpkey = nextkeyhashtable(bipartprobs))
  {
    counter = (double*)gethashtable(bipartprobs, bpkey);
    if ((*counter) < 2)      
      enqueuelinkedlist(removelist, bpkey);
      
    else
      *counter /= ntrees;
  }
  
  resetlinkedlist(removelist);
  
  while ((bpkey = dequeuelinkedlist(removelist)) != NULL)
  //{
    free(removehashtable(bipartprobs, bpkey));
    //freetreebipart((treebipart*)bpkey);
    //free(bpkey);
  //}
  
  freelinkedlist(removelist);
  free(removelist);
  
  return TRUE;
  
}

int main()
{
  int i;
  const int iterations = 100;
  const int ntaxa = 50;
  char treelist[22] = "gene000top00.treelist";
  char **taxa = (char**)malloc(ntaxa * sizeof(char*));
  bool genetaxa[ntaxa];
  hashtable *bp = (hashtable*)malloc(sizeof(hashtable));
  //inithashtable(bp, hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
  
  for (i = 0; i < ntaxa; i++)
  {
    taxa[i] = (char*)malloc(3 * sizeof(char));
    sprintf(taxa[i], "%d", i);
  }
  
  for (i = 0; i < iterations; i++)
  {
    inithashtable(bp, hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
    readtrees(bp, treelist, taxa, ntaxa);
    freehashtable(bp);
    /*inithashtable(bp, hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
    probbiparts(bp, treelist, taxa, ntaxa, genetaxa, 250);
    freehashtable(bp); */
  }
  
  for (i = 0; i < ntaxa; i++)
    free(taxa[i]);
  free(taxa);
  
  //freehashtable(bp);
  free(bp);

}